#    Copyright (C) 2014 Yahoo! Inc. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import networkx as nx
import pickle

from taskflow import test
from taskflow.types import graph
from taskflow.types import sets
from taskflow.types import timing
from taskflow.types import tree


class TimingTest(test.TestCase):
    def test_convert_fail(self):
        for baddie in ["abc123", "-1", "", object()]:
            self.assertRaises(ValueError,
                              timing.convert_to_timeout, baddie)

    def test_convert_noop(self):
        t = timing.convert_to_timeout(1.0)
        t2 = timing.convert_to_timeout(t)
        self.assertEqual(t, t2)

    def test_interrupt(self):
        t = timing.convert_to_timeout(1.0)
        self.assertFalse(t.is_stopped())
        t.interrupt()
        self.assertTrue(t.is_stopped())

    def test_reset(self):
        t = timing.convert_to_timeout(1.0)
        t.interrupt()
        self.assertTrue(t.is_stopped())
        t.reset()
        self.assertFalse(t.is_stopped())

    def test_values(self):
        for v, e_v in [("1.0", 1.0), (1, 1.0),
                       ("2.0", 2.0)]:
            t = timing.convert_to_timeout(v)
            self.assertEqual(e_v, t.value)

    def test_fail(self):
        self.assertRaises(ValueError,
                          timing.Timeout, -1)


class GraphTest(test.TestCase):
    def test_no_successors_no_predecessors(self):
        g = graph.DiGraph()
        g.add_node("a")
        g.add_node("b")
        g.add_node("c")
        g.add_edge("b", "c")
        self.assertEqual({'a', 'b'},
                         set(g.no_predecessors_iter()))
        self.assertEqual({'a', 'c'},
                         set(g.no_successors_iter()))

    def test_directed(self):
        g = graph.DiGraph()
        g.add_node("a")
        g.add_node("b")
        g.add_edge("a", "b")
        self.assertTrue(g.is_directed_acyclic())
        g.add_edge("b", "a")
        self.assertFalse(g.is_directed_acyclic())

    def test_frozen(self):
        g = graph.DiGraph()
        self.assertFalse(g.frozen)
        g.add_node("b")
        g.freeze()
        self.assertRaises(nx.NetworkXError, g.add_node, "c")

    def test_merge(self):
        g = graph.DiGraph()
        g.add_node("a")
        g.add_node("b")

        g2 = graph.DiGraph()
        g2.add_node('c')

        g3 = graph.merge_graphs(g, g2)
        self.assertEqual(3, len(g3))

    def test_pydot_output(self):
        # NOTE(harlowja): ensure we use the ordered types here, otherwise
        # the expected output will vary based on randomized hashing and then
        # the test will fail randomly...
        for graph_cls, kind, edge in [(graph.OrderedDiGraph, 'digraph', '->'),
                                      (graph.OrderedGraph, 'graph', '--')]:
            g = graph_cls(name='test')
            g.add_node("a")
            g.add_node("b")
            g.add_node("c")
            g.add_edge("a", "b")
            g.add_edge("b", "c")
            expected = """
strict %(kind)s "test" {
a;
b;
c;
a %(edge)s b;
b %(edge)s c;
}
""" % ({'kind': kind, 'edge': edge})
            self.assertEqual(expected.lstrip(), g.export_to_dot())

    def test_merge_edges(self):
        g = graph.DiGraph()
        g.add_node("a")
        g.add_node("b")
        g.add_edge('a', 'b')

        g2 = graph.DiGraph()
        g2.add_node('c')
        g2.add_node('d')
        g2.add_edge('c', 'd')

        g3 = graph.merge_graphs(g, g2)
        self.assertEqual(4, len(g3))
        self.assertTrue(g3.has_edge('c', 'd'))
        self.assertTrue(g3.has_edge('a', 'b'))

    def test_overlap_detector(self):
        g = graph.DiGraph()
        g.add_node("a")
        g.add_node("b")
        g.add_edge('a', 'b')

        g2 = graph.DiGraph()
        g2.add_node('a')
        g2.add_node('d')
        g2.add_edge('a', 'd')

        self.assertRaises(ValueError,
                          graph.merge_graphs, g, g2)

        def occurrence_detector(to_graph, from_graph):
            return sum(1 for node in from_graph.nodes if node in to_graph)

        self.assertRaises(ValueError,
                          graph.merge_graphs, g, g2,
                          overlap_detector=occurrence_detector)

        g3 = graph.merge_graphs(g, g2, allow_overlaps=True)
        self.assertEqual(3, len(g3))
        self.assertTrue(g3.has_edge('a', 'b'))
        self.assertTrue(g3.has_edge('a', 'd'))

    def test_invalid_detector(self):
        g = graph.DiGraph()
        g.add_node("a")

        g2 = graph.DiGraph()
        g2.add_node('c')

        self.assertRaises(ValueError,
                          graph.merge_graphs, g, g2,
                          overlap_detector='b')


class TreeTest(test.TestCase):
    def _make_species(self):
        # This is the following tree:
        #
        # animal
        # |__mammal
        # |  |__horse
        # |  |__primate
        # |     |__monkey
        # |     |__human
        # |__reptile
        a = tree.Node("animal")
        m = tree.Node("mammal")
        r = tree.Node("reptile")
        a.add(m)
        a.add(r)
        m.add(tree.Node("horse"))
        p = tree.Node("primate")
        m.add(p)
        p.add(tree.Node("monkey"))
        p.add(tree.Node("human"))
        return a

    def test_pformat_species(self):
        root = self._make_species()
        expected = """
animal
|__mammal
|  |__horse
|  |__primate
|     |__monkey
|     |__human
|__reptile
"""
        self.assertEqual(expected.strip(), root.pformat())

    def test_pformat_flat(self):
        root = tree.Node("josh")
        root.add(tree.Node("josh.1"))
        expected = """
josh
|__josh.1
"""
        self.assertEqual(expected.strip(), root.pformat())

        root[0].add(tree.Node("josh.1.1"))
        expected = """
josh
|__josh.1
   |__josh.1.1
"""
        self.assertEqual(expected.strip(), root.pformat())

        root[0][0].add(tree.Node("josh.1.1.1"))
        expected = """
josh
|__josh.1
   |__josh.1.1
      |__josh.1.1.1
"""
        self.assertEqual(expected.strip(), root.pformat())

        root[0][0][0].add(tree.Node("josh.1.1.1.1"))
        expected = """
josh
|__josh.1
   |__josh.1.1
      |__josh.1.1.1
         |__josh.1.1.1.1
"""
        self.assertEqual(expected.strip(), root.pformat())

    def test_pformat_partial_species(self):
        root = self._make_species()

        expected = """
reptile
"""
        self.assertEqual(expected.strip(), root[1].pformat())

        expected = """
mammal
|__horse
|__primate
   |__monkey
   |__human
"""
        self.assertEqual(expected.strip(), root[0].pformat())

        expected = """
primate
|__monkey
|__human
"""
        self.assertEqual(expected.strip(), root[0][1].pformat())

        expected = """
monkey
"""
        self.assertEqual(expected.strip(), root[0][1][0].pformat())

    def test_pformat(self):

        root = tree.Node("CEO")

        expected = """
CEO
"""

        self.assertEqual(expected.strip(), root.pformat())

        root.add(tree.Node("Infra"))

        expected = """
CEO
|__Infra
"""
        self.assertEqual(expected.strip(), root.pformat())

        root[0].add(tree.Node("Infra.1"))
        expected = """
CEO
|__Infra
   |__Infra.1
"""
        self.assertEqual(expected.strip(), root.pformat())

        root.add(tree.Node("Mail"))
        expected = """
CEO
|__Infra
|  |__Infra.1
|__Mail
"""
        self.assertEqual(expected.strip(), root.pformat())

        root.add(tree.Node("Search"))
        expected = """
CEO
|__Infra
|  |__Infra.1
|__Mail
|__Search
"""
        self.assertEqual(expected.strip(), root.pformat())

        root[-1].add(tree.Node("Search.1"))
        expected = """
CEO
|__Infra
|  |__Infra.1
|__Mail
|__Search
   |__Search.1
"""
        self.assertEqual(expected.strip(), root.pformat())

        root[-1].add(tree.Node("Search.2"))
        expected = """
CEO
|__Infra
|  |__Infra.1
|__Mail
|__Search
   |__Search.1
   |__Search.2
"""
        self.assertEqual(expected.strip(), root.pformat())

        root[0].add(tree.Node("Infra.2"))
        expected = """
CEO
|__Infra
|  |__Infra.1
|  |__Infra.2
|__Mail
|__Search
   |__Search.1
   |__Search.2
"""
        self.assertEqual(expected.strip(), root.pformat())

        root[0].add(tree.Node("Infra.3"))
        expected = """
CEO
|__Infra
|  |__Infra.1
|  |__Infra.2
|  |__Infra.3
|__Mail
|__Search
   |__Search.1
   |__Search.2
"""
        self.assertEqual(expected.strip(), root.pformat())

        root[0][-1].add(tree.Node("Infra.3.1"))
        expected = """
CEO
|__Infra
|  |__Infra.1
|  |__Infra.2
|  |__Infra.3
|     |__Infra.3.1
|__Mail
|__Search
   |__Search.1
   |__Search.2
"""
        self.assertEqual(expected.strip(), root.pformat())

        root[-1][0].add(tree.Node("Search.1.1"))
        expected = """
CEO
|__Infra
|  |__Infra.1
|  |__Infra.2
|  |__Infra.3
|     |__Infra.3.1
|__Mail
|__Search
   |__Search.1
   |  |__Search.1.1
   |__Search.2
"""
        self.assertEqual(expected.strip(), root.pformat())

        root[1].add(tree.Node("Mail.1"))
        expected = """
CEO
|__Infra
|  |__Infra.1
|  |__Infra.2
|  |__Infra.3
|     |__Infra.3.1
|__Mail
|  |__Mail.1
|__Search
   |__Search.1
   |  |__Search.1.1
   |__Search.2
"""
        self.assertEqual(expected.strip(), root.pformat())

        root[1][0].add(tree.Node("Mail.1.1"))
        expected = """
CEO
|__Infra
|  |__Infra.1
|  |__Infra.2
|  |__Infra.3
|     |__Infra.3.1
|__Mail
|  |__Mail.1
|     |__Mail.1.1
|__Search
   |__Search.1
   |  |__Search.1.1
   |__Search.2
"""
        self.assertEqual(expected.strip(), root.pformat())

    def test_path(self):
        root = self._make_species()
        human = root.find("human")
        self.assertIsNotNone(human)
        p = list([n.item for n in human.path_iter()])
        self.assertEqual(['human', 'primate', 'mammal', 'animal'], p)

    def test_empty(self):
        root = tree.Node("josh")
        self.assertTrue(root.empty())

    def test_after_frozen(self):
        root = tree.Node("josh")
        root.add(tree.Node("josh.1"))
        root.freeze()
        self.assertTrue(
            all(n.frozen for n in root.dfs_iter(include_self=True)))
        self.assertRaises(tree.FrozenNode,
                          root.remove, "josh.1")
        self.assertRaises(tree.FrozenNode, root.disassociate)
        self.assertRaises(tree.FrozenNode, root.add,
                          tree.Node("josh.2"))

    def test_removal(self):
        root = self._make_species()
        self.assertIsNotNone(root.remove('reptile'))
        self.assertRaises(ValueError, root.remove, 'reptile')
        self.assertIsNone(root.find('reptile'))

    def test_removal_direct(self):
        root = self._make_species()
        self.assertRaises(ValueError, root.remove, 'human',
                          only_direct=True)

    def test_removal_self(self):
        root = self._make_species()
        n = root.find('horse')
        self.assertIsNotNone(n.parent)
        n.remove('horse', include_self=True)
        self.assertIsNone(n.parent)
        self.assertIsNone(root.find('horse'))

    def test_disassociate(self):
        root = self._make_species()
        n = root.find('horse')
        self.assertIsNotNone(n.parent)
        c = n.disassociate()
        self.assertEqual(1, c)
        self.assertIsNone(n.parent)
        self.assertIsNone(root.find('horse'))

    def test_disassociate_many(self):
        root = self._make_species()
        n = root.find('horse')
        n.parent.add(n)
        n.parent.add(n)
        c = n.disassociate()
        self.assertEqual(3, c)
        self.assertIsNone(n.parent)
        self.assertIsNone(root.find('horse'))

    def test_not_empty(self):
        root = self._make_species()
        self.assertFalse(root.empty())

    def test_node_count(self):
        root = self._make_species()
        self.assertEqual(7, 1 + root.child_count(only_direct=False))

    def test_index(self):
        root = self._make_species()
        self.assertEqual(0, root.index("mammal"))
        self.assertEqual(1, root.index("reptile"))

    def test_contains(self):
        root = self._make_species()
        self.assertIn("monkey", root)
        self.assertNotIn("bird", root)

    def test_freeze(self):
        root = self._make_species()
        root.freeze()
        self.assertRaises(tree.FrozenNode, root.add, "bird")

    def test_find(self):
        root = self._make_species()
        self.assertIsNone(root.find('monkey', only_direct=True))
        self.assertIsNotNone(root.find('monkey', only_direct=False))
        self.assertIsNotNone(root.find('animal', only_direct=True))
        self.assertIsNotNone(root.find('reptile', only_direct=True))
        self.assertIsNone(root.find('animal', include_self=False))
        self.assertIsNone(root.find('animal',
                                    include_self=False, only_direct=True))

    def test_dfs_itr(self):
        root = self._make_species()
        things = list([n.item for n in root.dfs_iter(include_self=True)])
        self.assertEqual({'animal', 'reptile', 'mammal', 'horse',
                          'primate', 'monkey', 'human'}, set(things))

    def test_dfs_itr_left_to_right(self):
        root = self._make_species()
        it = root.dfs_iter(include_self=False, right_to_left=False)
        things = list([n.item for n in it])
        self.assertEqual(['reptile', 'mammal', 'primate',
                          'human', 'monkey', 'horse'], things)

    def test_dfs_itr_no_self(self):
        root = self._make_species()
        things = list([n.item for n in root.dfs_iter(include_self=False)])
        self.assertEqual(['mammal', 'horse', 'primate',
                          'monkey', 'human', 'reptile'], things)

    def test_bfs_itr(self):
        root = self._make_species()
        things = list([n.item for n in root.bfs_iter(include_self=True)])
        self.assertEqual(['animal', 'reptile', 'mammal', 'primate',
                          'horse', 'human', 'monkey'], things)

    def test_bfs_itr_no_self(self):
        root = self._make_species()
        things = list([n.item for n in root.bfs_iter(include_self=False)])
        self.assertEqual(['reptile', 'mammal', 'primate',
                          'horse', 'human', 'monkey'], things)

    def test_bfs_itr_right_to_left(self):
        root = self._make_species()
        it = root.bfs_iter(include_self=False, right_to_left=True)
        things = list([n.item for n in it])
        self.assertEqual(['mammal', 'reptile', 'horse',
                          'primate', 'monkey', 'human'], things)

    def test_to_diagraph(self):
        root = self._make_species()
        g = root.to_digraph()
        self.assertEqual(root.child_count(only_direct=False) + 1, len(g))
        for node in root.dfs_iter(include_self=True):
            self.assertIn(node.item, g)
        self.assertEqual([], list(g.predecessors('animal')))
        self.assertEqual(['animal'], list(g.predecessors('reptile')))
        self.assertEqual(['primate'], list(g.predecessors('human')))
        self.assertEqual(['mammal'], list(g.predecessors('primate')))
        self.assertEqual(['animal'], list(g.predecessors('mammal')))
        self.assertEqual(['mammal', 'reptile'], list(g.successors('animal')))

    def test_to_digraph_retains_metadata(self):
        root = tree.Node("chickens", alive=True)
        dead_chicken = tree.Node("chicken.1", alive=False)
        root.add(dead_chicken)
        g = root.to_digraph()
        self.assertEqual(g.nodes['chickens'], {'alive': True})
        self.assertEqual(g.nodes['chicken.1'], {'alive': False})


class OrderedSetTest(test.TestCase):

    def test_pickleable(self):
        items = [10, 9, 8, 7]
        s = sets.OrderedSet(items)
        self.assertEqual(items, list(s))
        s_bin = pickle.dumps(s)
        s2 = pickle.loads(s_bin)
        self.assertEqual(s, s2)
        self.assertEqual(items, list(s2))

    def test_retain_ordering(self):
        items = [10, 9, 8, 7]
        s = sets.OrderedSet(iter(items))
        self.assertEqual(items, list(s))

    def test_retain_duplicate_ordering(self):
        items = [10, 9, 10, 8, 9, 7, 8]
        s = sets.OrderedSet(iter(items))
        self.assertEqual([10, 9, 8, 7], list(s))

    def test_length(self):
        items = [10, 9, 8, 7]
        s = sets.OrderedSet(iter(items))
        self.assertEqual(4, len(s))

    def test_duplicate_length(self):
        items = [10, 9, 10, 8, 9, 7, 8]
        s = sets.OrderedSet(iter(items))
        self.assertEqual(4, len(s))

    def test_contains(self):
        items = [10, 9, 8, 7]
        s = sets.OrderedSet(iter(items))
        for i in items:
            self.assertIn(i, s)

    def test_copy(self):
        items = [10, 9, 8, 7]
        s = sets.OrderedSet(iter(items))
        s2 = s.copy()
        self.assertEqual(s, s2)
        self.assertEqual(items, list(s2))

    def test_empty_intersection(self):
        s = sets.OrderedSet([1, 2, 3])

        es = set(s)

        self.assertEqual(es.intersection(), s.intersection())

    def test_intersection(self):
        s = sets.OrderedSet([1, 2, 3])
        s2 = sets.OrderedSet([2, 3, 4, 5])

        es = set(s)
        es2 = set(s2)

        self.assertEqual(es.intersection(es2), s.intersection(s2))
        self.assertEqual(es2.intersection(s), s2.intersection(s))

    def test_multi_intersection(self):
        s = sets.OrderedSet([1, 2, 3])
        s2 = sets.OrderedSet([2, 3, 4, 5])
        s3 = sets.OrderedSet([1, 2])

        es = set(s)
        es2 = set(s2)
        es3 = set(s3)

        self.assertEqual(es.intersection(s2, s3), s.intersection(s2, s3))
        self.assertEqual(es2.intersection(es3), s2.intersection(s3))

    def test_superset(self):
        s = sets.OrderedSet([1, 2, 3])
        s2 = sets.OrderedSet([2, 3])
        self.assertTrue(s.issuperset(s2))
        self.assertFalse(s.issubset(s2))

    def test_subset(self):
        s = sets.OrderedSet([1, 2, 3])
        s2 = sets.OrderedSet([2, 3])
        self.assertTrue(s2.issubset(s))
        self.assertFalse(s2.issuperset(s))

    def test_empty_difference(self):
        s = sets.OrderedSet([1, 2, 3])

        es = set(s)

        self.assertEqual(es.difference(), s.difference())

    def test_difference(self):
        s = sets.OrderedSet([1, 2, 3])
        s2 = sets.OrderedSet([2, 3])

        es = set(s)
        es2 = set(s2)

        self.assertEqual(es.difference(es2), s.difference(s2))
        self.assertEqual(es2.difference(es), s2.difference(s))

    def test_multi_difference(self):
        s = sets.OrderedSet([1, 2, 3])
        s2 = sets.OrderedSet([2, 3])
        s3 = sets.OrderedSet([3, 4, 5])

        es = set(s)
        es2 = set(s2)
        es3 = set(s3)

        self.assertEqual(es3.difference(es), s3.difference(s))
        self.assertEqual(es.difference(es3), s.difference(s3))
        self.assertEqual(es2.difference(es, es3), s2.difference(s, s3))

    def test_empty_union(self):
        s = sets.OrderedSet([1, 2, 3])

        es = set(s)

        self.assertEqual(es.union(), s.union())

    def test_union(self):
        s = sets.OrderedSet([1, 2, 3])
        s2 = sets.OrderedSet([2, 3, 4])

        es = set(s)
        es2 = set(s2)

        self.assertEqual(es.union(es2), s.union(s2))
        self.assertEqual(es2.union(es), s2.union(s))

    def test_multi_union(self):
        s = sets.OrderedSet([1, 2, 3])
        s2 = sets.OrderedSet([2, 3, 4])
        s3 = sets.OrderedSet([4, 5, 6])

        es = set(s)
        es2 = set(s2)
        es3 = set(s3)

        self.assertEqual(es.union(es2, es3), s.union(s2, s3))



#    Copyright (C) 2015 Yahoo! Inc. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging
import os
import sys

logging.basicConfig(level=logging.DEBUG)

top_dir = os.path.abspath(os.path.join(os.path.dirname(__file__),
                                       os.pardir,
                                       os.pardir))
sys.path.insert(0, top_dir)

from taskflow import engines
from taskflow.listeners import logging as logging_listener
from taskflow.patterns import linear_flow as lf
from taskflow import task

# INTRO: This example walks through a miniature workflow which will do a
# simple echo operation; during this execution a listener is associated with
# the engine to receive all notifications about what the flow has performed,
# this example dumps that output to the stdout for viewing (at debug level
# to show all the information which is possible).


class Echo(task.Task):
    def execute(self):
        print(self.name)


# Generate the work to be done (but don't do it yet).
wf = lf.Flow('abc')
wf.add(Echo('a'))
wf.add(Echo('b'))
wf.add(Echo('c'))

# This will associate the listener with the engine (the listener
# will automatically register for notifications with the engine and deregister
# when the context is exited).
e = engines.load(wf)
with logging_listener.DynamicLoggingListener(e):
    e.run()



#    Copyright (C) 2013 Yahoo! Inc. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import time
from unittest import mock

from oslo_utils import uuidutils
import testtools

from taskflow import exceptions as excp
from taskflow.jobs.backends import impl_redis
from taskflow import states
from taskflow import test
from taskflow.tests.unit.jobs import base
from taskflow.tests import utils as test_utils
from taskflow.utils import persistence_utils as p_utils
from taskflow.utils import redis_utils as ru


REDIS_AVAILABLE = test_utils.redis_available(
    impl_redis.RedisJobBoard.MIN_REDIS_VERSION)
REDIS_PORT = test_utils.REDIS_PORT


@testtools.skipIf(not REDIS_AVAILABLE, 'redis is not available')
class RedisJobboardTest(test.TestCase, base.BoardTestMixin):
    def close_client(self, client):
        client.close()

    def create_board(self, persistence=None):
        namespace = uuidutils.generate_uuid()
        client = ru.RedisClient(port=REDIS_PORT)
        config = {
            'namespace': ("taskflow-%s" % namespace).encode('latin-1'),
        }
        kwargs = {
            'client': client,
            'persistence': persistence,
        }
        board = impl_redis.RedisJobBoard('test-board', config, **kwargs)
        self.addCleanup(board.close)
        self.addCleanup(self.close_client, client)
        return (client, board)

    def test_posting_claim_expiry(self):

        with base.connect_close(self.board):
            with self.flush(self.client):
                self.board.post('test', p_utils.temporary_log_book())

            self.assertEqual(1, self.board.job_count)
            possible_jobs = list(self.board.iterjobs(only_unclaimed=True))
            self.assertEqual(1, len(possible_jobs))
            j = possible_jobs[0]
            self.assertEqual(states.UNCLAIMED, j.state)

            with self.flush(self.client):
                self.board.claim(j, self.board.name, expiry=0.5)

            self.assertEqual(self.board.name, self.board.find_owner(j))
            self.assertEqual(states.CLAIMED, j.state)

            time.sleep(0.6)
            self.assertEqual(states.UNCLAIMED, j.state)
            possible_jobs = list(self.board.iterjobs(only_unclaimed=True))
            self.assertEqual(1, len(possible_jobs))

    def test_posting_claim_same_owner(self):
        with base.connect_close(self.board):
            with self.flush(self.client):
                self.board.post('test', p_utils.temporary_log_book())

            self.assertEqual(1, self.board.job_count)
            possible_jobs = list(self.board.iterjobs(only_unclaimed=True))
            self.assertEqual(1, len(possible_jobs))
            j = possible_jobs[0]
            self.assertEqual(states.UNCLAIMED, j.state)

            with self.flush(self.client):
                self.board.claim(j, self.board.name)

            possible_jobs = list(self.board.iterjobs())
            self.assertEqual(1, len(possible_jobs))
            with self.flush(self.client):
                self.assertRaises(excp.UnclaimableJob, self.board.claim,
                                  possible_jobs[0], self.board.name)
            possible_jobs = list(self.board.iterjobs(only_unclaimed=True))
            self.assertEqual(0, len(possible_jobs))

    def setUp(self):
        super().setUp()
        self.client, self.board = self.create_board()

    def test__make_client(self):
        conf = {'host': '127.0.0.1',
                'port': 6379,
                'username': 'default',
                'password': 'secret',
                'namespace': 'test'
                }
        test_conf = {
            'host': '127.0.0.1',
            'port': 6379,
            'username': 'default',
            'password': 'secret',
        }
        with mock.patch('taskflow.utils.redis_utils.RedisClient') as mock_ru:
            impl_redis.RedisJobBoard('test-board', conf)
            mock_ru.assert_called_once_with(**test_conf)

    def test__make_client_sentinel(self):
        conf = {'host': '127.0.0.1',
                'port': 26379,
                'username': 'default',
                'password': 'secret',
                'namespace': 'test',
                'sentinel': 'mymaster',
                'sentinel_kwargs': {
                    'username': 'default',
                    'password': 'senitelsecret'
                }}
        with mock.patch('redis.sentinel.Sentinel') as mock_sentinel:
            impl_redis.RedisJobBoard('test-board', conf)
            test_conf = {
                'username': 'default',
                'password': 'secret',
            }
            mock_sentinel.assert_called_once_with(
                [('127.0.0.1', 26379)],
                sentinel_kwargs={
                    'username': 'default',
                    'password': 'senitelsecret'
                },
                **test_conf)
            mock_sentinel().master_for.assert_called_once_with('mymaster')

    def test__make_client_sentinel_fallbacks(self):
        conf = {'host': '127.0.0.1',
                'port': 26379,
                'username': 'default',
                'password': 'secret',
                'namespace': 'test',
                'sentinel': 'mymaster',
                'sentinel_fallbacks': [
                    '[::1]:26379', '127.0.0.2:26379', 'localhost:26379'
                ]}
        with mock.patch('redis.sentinel.Sentinel') as mock_sentinel:
            impl_redis.RedisJobBoard('test-board', conf)
            test_conf = {
                'username': 'default',
                'password': 'secret',
                'sentinel_kwargs': None,
            }
            mock_sentinel.assert_called_once_with(
                [('127.0.0.1', 26379), ('::1', 26379),
                 ('127.0.0.2', 26379), ('localhost', 26379)],
                **test_conf)
            mock_sentinel().master_for.assert_called_once_with('mymaster')

    def test__make_client_sentinel_ssl(self):
        conf = {'host': '127.0.0.1',
                'port': 26379,
                'username': 'default',
                'password': 'secret',
                'namespace': 'test',
                'sentinel': 'mymaster',
                'sentinel_kwargs': None,
                'ssl': True,
                'ssl_ca_certs': '/etc/ssl/certs'}
        with mock.patch('redis.sentinel.Sentinel') as mock_sentinel:
            impl_redis.RedisJobBoard('test-board', conf)
            test_conf = {
                'username': 'default',
                'password': 'secret',
                'ssl': True,
                'ssl_ca_certs': '/etc/ssl/certs',
            }
            mock_sentinel.assert_called_once_with(
                [('127.0.0.1', 26379)],
                sentinel_kwargs=None,
                **test_conf)
            mock_sentinel().master_for.assert_called_once_with('mymaster')



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Add unique into all indexes

Revision ID: 00af93df9d77
Revises: 40fc8c914bd2
Create Date: 2025-02-28 15:44:37.066720

"""

# revision identifiers, used by Alembic.
revision = '00af93df9d77'
down_revision = '40fc8c914bd2'

from alembic import op


def upgrade():
    bind = op.get_bind()
    engine = bind.engine
    if engine.name == 'mysql':
        with op.batch_alter_table("logbooks") as batch_op:
            batch_op.drop_index("logbook_uuid_idx")
            batch_op.create_index(
                index_name="logbook_uuid_idx",
                columns=['uuid'],
                unique=True)

        with op.batch_alter_table("flowdetails") as batch_op:
            batch_op.drop_index("flowdetails_uuid_idx")
            batch_op.create_index(
                index_name="flowdetails_uuid_idx",
                columns=['uuid'],
                unique=True)

        with op.batch_alter_table("atomdetails") as batch_op:
            batch_op.drop_index("taskdetails_uuid_idx")
            batch_op.create_index(
                index_name="taskdetails_uuid_idx",
                columns=['uuid'],
                unique=True)


def downgrade():
    bind = op.get_bind()
    engine = bind.engine
    if engine.name == 'mysql':
        with op.batch_alter_table("logbooks") as batch_op:
            batch_op.drop_index("logbook_uuid_idx")
            batch_op.create_index(
                index_name="logbook_uuid_idx",
                columns=['uuid'])

        with op.batch_alter_table("flowdetails") as batch_op:
            batch_op.drop_index("flowdetails_uuid_idx")
            batch_op.create_index(
                index_name="flowdetails_uuid_idx",
                columns=['uuid'])

        with op.batch_alter_table("atomdetails") as batch_op:
            batch_op.drop_index("taskdetails_uuid_idx")
            batch_op.create_index(
                index_name="taskdetails_uuid_idx",
                columns=['uuid'])



#    Copyright (C) 2014 Yahoo! Inc. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging
import os
import random
import sys
import time

logging.basicConfig(level=logging.ERROR)

top_dir = os.path.abspath(os.path.join(os.path.dirname(__file__),
                                       os.pardir,
                                       os.pardir))
sys.path.insert(0, top_dir)

from taskflow import engines
from taskflow.listeners import timing
from taskflow.patterns import linear_flow as lf
from taskflow import task

# INTRO: in this example we will attach a listener to an engine
# and have variable run time tasks run and show how the listener will print
# out how long those tasks took (when they started and when they finished).
#
# This shows how timing metrics can be gathered (or attached onto an engine)
# after a workflow has been constructed, making it easy to gather metrics
# dynamically for situations where this kind of information is applicable (or
# even adding this information on at a later point in the future when your
# application starts to slow down).


class VariableTask(task.Task):
    def __init__(self, name):
        super().__init__(name)
        self._sleepy_time = random.random()

    def execute(self):
        time.sleep(self._sleepy_time)


f = lf.Flow('root')
f.add(VariableTask('a'), VariableTask('b'), VariableTask('c'))
e = engines.load(f)
with timing.PrintingDurationListener(e):
    e.run()



#    Copyright (C) 2014 Yahoo! Inc. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from taskflow.engines.worker_based import engine
from taskflow.engines.worker_based import executor
from taskflow.patterns import linear_flow as lf
from taskflow.persistence import backends
from taskflow import test
from taskflow.test import mock
from taskflow.tests import utils
from taskflow.utils import persistence_utils as pu


class TestWorkerBasedActionEngine(test.MockTestCase):
    @staticmethod
    def _create_engine(**kwargs):
        flow = lf.Flow('test-flow').add(utils.DummyTask())
        backend = backends.fetch({'connection': 'memory'})
        flow_detail = pu.create_flow_detail(flow, backend=backend)
        options = kwargs.copy()
        return engine.WorkerBasedActionEngine(flow, flow_detail,
                                              backend, options)

    def _patch_in_executor(self):
        executor_mock, executor_inst_mock = self.patchClass(
            engine.executor, 'WorkerTaskExecutor', attach_as='executor')
        return executor_mock, executor_inst_mock

    def test_creation_default(self):
        executor_mock, executor_inst_mock = self._patch_in_executor()
        eng = self._create_engine()
        expected_calls = [
            mock.call.executor_class(uuid=eng.storage.flow_uuid,
                                     url=None,
                                     exchange='default',
                                     topics=[],
                                     transport=None,
                                     transport_options=None,
                                     transition_timeout=mock.ANY,
                                     retry_options=None,
                                     worker_expiry=mock.ANY)
        ]
        self.assertEqual(expected_calls, self.master_mock.mock_calls)

    def test_creation_custom(self):
        executor_mock, executor_inst_mock = self._patch_in_executor()
        topics = ['test-topic1', 'test-topic2']
        exchange = 'test-exchange'
        broker_url = 'test-url'
        eng = self._create_engine(
            url=broker_url,
            exchange=exchange,
            transport='memory',
            transport_options={},
            transition_timeout=200,
            topics=topics,
            retry_options={},
            worker_expiry=1)
        expected_calls = [
            mock.call.executor_class(uuid=eng.storage.flow_uuid,
                                     url=broker_url,
                                     exchange=exchange,
                                     topics=topics,
                                     transport='memory',
                                     transport_options={},
                                     transition_timeout=200,
                                     retry_options={},
                                     worker_expiry=1)
        ]
        self.assertEqual(expected_calls, self.master_mock.mock_calls)

    def test_creation_custom_executor(self):
        ex = executor.WorkerTaskExecutor('a', 'test-exchange', ['test-topic'])
        eng = self._create_engine(executor=ex)
        self.assertIs(eng._task_executor, ex)
        self.assertIsInstance(eng._task_executor, executor.WorkerTaskExecutor)

    def test_creation_invalid_custom_executor(self):
        self.assertRaises(TypeError, self._create_engine, executor=2)
        self.assertRaises(TypeError, self._create_engine, executor='blah')



#    Copyright (C) 2014 Yahoo! Inc. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import socket
from unittest import mock

from taskflow.engines.worker_based import proxy
from taskflow import test
from taskflow.utils import threading_utils


class TestProxy(test.MockTestCase):

    def setUp(self):
        super().setUp()
        self.topic = 'test-topic'
        self.broker_url = 'test-url'
        self.exchange = 'test-exchange'
        self.timeout = 5
        self.de_period = proxy.DRAIN_EVENTS_PERIOD

        # patch classes
        self.conn_mock, self.conn_inst_mock = self.patchClass(
            proxy.kombu, 'Connection')
        self.exchange_mock, self.exchange_inst_mock = self.patchClass(
            proxy.kombu, 'Exchange')
        self.queue_mock, self.queue_inst_mock = self.patchClass(
            proxy.kombu, 'Queue')
        self.producer_mock, self.producer_inst_mock = self.patchClass(
            proxy.kombu, 'Producer')

        # connection mocking
        def _ensure(obj, func, *args, **kwargs):
            return func
        self.conn_inst_mock.drain_events.side_effect = [
            socket.timeout, socket.timeout, KeyboardInterrupt]
        self.conn_inst_mock.ensure = mock.MagicMock(side_effect=_ensure)

        # connections mocking
        self.connections_mock = self.patch(
            "taskflow.engines.worker_based.proxy.kombu.connections",
            attach_as='connections')
        self.connections_mock.__getitem__().acquire().__enter__.return_value =\
            self.conn_inst_mock

        # producers mocking
        self.conn_inst_mock.Producer.return_value.__enter__ = mock.MagicMock()
        self.conn_inst_mock.Producer.return_value.__exit__ = mock.MagicMock()

        # consumer mocking
        self.conn_inst_mock.Consumer.return_value.__enter__ = mock.MagicMock()
        self.conn_inst_mock.Consumer.return_value.__exit__ = mock.MagicMock()

        # other mocking
        self.on_wait_mock = mock.MagicMock(name='on_wait')
        self.master_mock.attach_mock(self.on_wait_mock, 'on_wait')

        # reset master mock
        self.resetMasterMock()

    def _queue_name(self, topic):
        return f"{self.exchange}_{topic}"

    def proxy_start_calls(self, calls, exc_type=mock.ANY):
        return [
            mock.call.Queue(name=self._queue_name(self.topic),
                            exchange=self.exchange_inst_mock,
                            routing_key=self.topic,
                            durable=False,
                            auto_delete=True,
                            channel=self.conn_inst_mock),
            mock.call.connection.Consumer(queues=self.queue_inst_mock,
                                          callbacks=[mock.ANY]),
            mock.call.connection.Consumer().__enter__(),
            mock.call.connection.ensure(mock.ANY, mock.ANY,
                                        interval_start=mock.ANY,
                                        interval_max=mock.ANY,
                                        max_retries=mock.ANY,
                                        interval_step=mock.ANY,
                                        errback=mock.ANY),
        ] + calls + [
            mock.call.connection.Consumer().__exit__(exc_type, mock.ANY,
                                                     mock.ANY)
        ]

    def proxy_publish_calls(self, calls, routing_key, exc_type=mock.ANY):
        return [
            mock.call.connection.Producer(),
            mock.call.connection.Producer().__enter__(),
            mock.call.connection.ensure(mock.ANY, mock.ANY,
                                        interval_start=mock.ANY,
                                        interval_max=mock.ANY,
                                        max_retries=mock.ANY,
                                        interval_step=mock.ANY,
                                        errback=mock.ANY),
            mock.call.Queue(name=self._queue_name(routing_key),
                            routing_key=routing_key,
                            exchange=self.exchange_inst_mock,
                            durable=False,
                            auto_delete=True,
                            channel=None),
        ] + calls + [
            mock.call.connection.Producer().__exit__(exc_type, mock.ANY,
                                                     mock.ANY)
        ]

    def proxy(self, reset_master_mock=False, **kwargs):
        proxy_kwargs = dict(topic=self.topic,
                            exchange=self.exchange,
                            url=self.broker_url,
                            type_handlers={})
        proxy_kwargs.update(kwargs)
        p = proxy.Proxy(**proxy_kwargs)
        if reset_master_mock:
            self.resetMasterMock()
        return p

    def test_creation(self):
        self.proxy()

        master_mock_calls = [
            mock.call.Connection(self.broker_url, transport=None,
                                 transport_options=None),
            mock.call.Exchange(name=self.exchange,
                               durable=False,
                               auto_delete=True)
        ]
        self.assertEqual(master_mock_calls, self.master_mock.mock_calls)

    def test_creation_custom(self):
        transport_opts = {'context': 'context'}
        self.proxy(transport='memory', transport_options=transport_opts)

        master_mock_calls = [
            mock.call.Connection(self.broker_url, transport='memory',
                                 transport_options=transport_opts),
            mock.call.Exchange(name=self.exchange,
                               durable=False,
                               auto_delete=True)
        ]
        self.assertEqual(master_mock_calls, self.master_mock.mock_calls)

    def test_publish(self):
        msg_mock = mock.MagicMock()
        msg_data = 'msg-data'
        msg_mock.to_dict.return_value = msg_data
        routing_key = 'routing-key'
        task_uuid = 'task-uuid'

        p = self.proxy(reset_master_mock=True)
        p.publish(msg_mock, routing_key, correlation_id=task_uuid)

        mock_producer = mock.call.connection.Producer()
        master_mock_calls = self.proxy_publish_calls([
            mock_producer.__enter__().publish(body=msg_data,
                                              routing_key=routing_key,
                                              exchange=self.exchange_inst_mock,
                                              correlation_id=task_uuid,
                                              declare=[self.queue_inst_mock],
                                              type=msg_mock.TYPE,
                                              reply_to=None)
        ], routing_key)
        self.master_mock.assert_has_calls(master_mock_calls)

    def test_start(self):
        try:
            # KeyboardInterrupt will be raised after two iterations
            self.proxy(reset_master_mock=True).start()
        except KeyboardInterrupt:
            pass

        master_calls = self.proxy_start_calls([
            mock.call.connection.drain_events(timeout=self.de_period),
            mock.call.connection.drain_events(timeout=self.de_period),
            mock.call.connection.drain_events(timeout=self.de_period),
        ], exc_type=KeyboardInterrupt)
        self.master_mock.assert_has_calls(master_calls)

    def test_start_with_on_wait(self):
        try:
            # KeyboardInterrupt will be raised after two iterations
            self.proxy(reset_master_mock=True,
                       on_wait=self.on_wait_mock).start()
        except KeyboardInterrupt:
            pass

        master_calls = self.proxy_start_calls([
            mock.call.connection.drain_events(timeout=self.de_period),
            mock.call.on_wait(),
            mock.call.connection.drain_events(timeout=self.de_period),
            mock.call.on_wait(),
            mock.call.connection.drain_events(timeout=self.de_period),
        ], exc_type=KeyboardInterrupt)
        self.master_mock.assert_has_calls(master_calls)

    def test_start_with_on_wait_raises(self):
        self.on_wait_mock.side_effect = RuntimeError('Woot!')
        try:
            # KeyboardInterrupt will be raised after two iterations
            self.proxy(reset_master_mock=True,
                       on_wait=self.on_wait_mock).start()
        except KeyboardInterrupt:
            pass

        master_calls = self.proxy_start_calls([
            mock.call.connection.drain_events(timeout=self.de_period),
            mock.call.on_wait(),
        ], exc_type=RuntimeError)
        self.master_mock.assert_has_calls(master_calls)

    def test_stop(self):
        self.conn_inst_mock.drain_events.side_effect = socket.timeout

        # create proxy
        pr = self.proxy(reset_master_mock=True)

        # check that proxy is not running yes
        self.assertFalse(pr.is_running)

        # start proxy in separate thread
        t = threading_utils.daemon_thread(pr.start)
        t.start()

        # make sure proxy is started
        pr.wait()

        # check that proxy is running now
        self.assertTrue(pr.is_running)

        # stop proxy and wait for thread to finish
        pr.stop()

        # wait for thread to finish
        t.join()

        self.assertFalse(pr.is_running)



#    Copyright (C) 2014 Yahoo! Inc. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging
import os

from taskflow import exceptions
from taskflow.listeners import base
from taskflow import states

LOG = logging.getLogger(__name__)


class CheckingClaimListener(base.Listener):
    """Listener that interacts [engine, job, jobboard]; ensures claim is valid.

    This listener (or a derivative) can be associated with an engines
    notification system after the job has been claimed (so that the jobs work
    can be worked on by that engine). This listener (after associated) will
    check that the job is still claimed *whenever* the engine notifies of a
    task or flow state change. If the job is not claimed when a state change
    occurs, a associated handler (or the default) will be activated to
    determine how to react to this *hopefully* exceptional case.

    NOTE(harlowja): this may create more traffic than desired to the
    jobboard backend (zookeeper or other), since the amount of state change
    per task and flow is non-zero (and checking during each state change will
    result in quite a few calls to that management system to check the jobs
    claim status); this could be later optimized to check less (or only check
    on a smaller set of states)

    NOTE(harlowja): if a custom ``on_job_loss`` callback is provided it must
    accept three positional arguments, the first being the current engine being
    ran, the second being the 'task/flow' state and the third being the details
    that were sent from the engine to listeners for inspection.
    """

    def __init__(self, engine, job, board, owner, on_job_loss=None):
        super().__init__(engine)
        self._job = job
        self._board = board
        self._owner = owner
        if on_job_loss is None:
            self._on_job_loss = self._suspend_engine_on_loss
        else:
            if not callable(on_job_loss):
                raise ValueError("Custom 'on_job_loss' handler must be"
                                 " callable")
            self._on_job_loss = on_job_loss

    def _suspend_engine_on_loss(self, engine, state, details):
        """The default strategy for handling claims being lost."""
        try:
            engine.suspend()
        except exceptions.TaskFlowException as e:
            LOG.warning("Failed suspending engine '%s', (previously owned by"
                        " '%s'):%s%s", engine, self._owner, os.linesep,
                        e.pformat())

    def _flow_receiver(self, state, details):
        self._claim_checker(state, details)

    def _task_receiver(self, state, details):
        self._claim_checker(state, details)

    def _has_been_lost(self):
        try:
            job_state = self._job.state
            job_owner = self._board.find_owner(self._job)
        except (exceptions.NotFound, exceptions.JobFailure):
            return True
        else:
            if job_state == states.UNCLAIMED or self._owner != job_owner:
                return True
            else:
                return False

    def _claim_checker(self, state, details):
        if not self._has_been_lost():
            LOG.debug("Job '%s' is still claimed (actively owned by '%s')",
                      self._job, self._owner)
        else:
            LOG.warning("Job '%s' has lost its claim"
                        " (previously owned by '%s')",
                        self._job, self._owner)
            self._on_job_loss(self._engine, state, details)



#    Copyright (C) 2014 Yahoo! Inc. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging
import os
import sys

logging.basicConfig(level=logging.ERROR)

top_dir = os.path.abspath(os.path.join(os.path.dirname(__file__),
                                       os.pardir,
                                       os.pardir))
sys.path.insert(0, top_dir)

from taskflow import engines
from taskflow.patterns import linear_flow as lf
from taskflow.patterns import unordered_flow as uf
from taskflow import task


# INTRO: This is the defacto hello world equivalent for taskflow; it shows how
# an overly simplistic workflow can be created that runs using different
# engines using different styles of execution (all can be used to run in
# parallel if a workflow is provided that is parallelizable).

class PrinterTask(task.Task):
    def __init__(self, name, show_name=True, inject=None):
        super().__init__(name, inject=inject)
        self._show_name = show_name

    def execute(self, output):
        if self._show_name:
            print(f"{self.name}: {output}")
        else:
            print(output)


# This will be the work that we want done, which for this example is just to
# print 'hello world' (like a song) using different tasks and different
# execution models.
song = lf.Flow("beats")

# Unordered flows when ran can be ran in parallel; and a chorus is everyone
# singing at once of course!
hi_chorus = uf.Flow('hello')
world_chorus = uf.Flow('world')
for (name, hello, world) in [('bob', 'hello', 'world'),
                             ('joe', 'hellooo', 'worllllld'),
                             ('sue', "helloooooo!", 'wooorllld!')]:
    hi_chorus.add(PrinterTask("%s@hello" % name,
                              # This will show up to the execute() method of
                              # the task as the argument named 'output' (which
                              # will allow us to print the character we want).
                              inject={'output': hello}))
    world_chorus.add(PrinterTask("%s@world" % name,
                                 inject={'output': world}))

# The composition starts with the conductor and then runs in sequence with
# the chorus running in parallel, but no matter what the 'hello' chorus must
# always run before the 'world' chorus (otherwise the world will fall apart).
song.add(PrinterTask("conductor@begin",
                     show_name=False, inject={'output': "*ding*"}),
         hi_chorus,
         world_chorus,
         PrinterTask("conductor@end",
                     show_name=False, inject={'output': "*dong*"}))

# Run in parallel using eventlet green threads...
try:
    import eventlet as _eventlet  # noqa
except ImportError:
    # No eventlet currently active, skip running with it...
    pass
else:
    print("-- Running in parallel using eventlet --")
    e = engines.load(song, executor='greenthreaded', engine='parallel',
                     max_workers=1)
    e.run()


# Run in parallel using real threads...
print("-- Running in parallel using threads --")
e = engines.load(song, executor='threaded', engine='parallel',
                 max_workers=1)
e.run()


# Run in parallel using external processes...
print("-- Running in parallel using processes --")
e = engines.load(song, executor='processes', engine='parallel',
                 max_workers=1)
e.run()


# Run serially (aka, if the workflow could have been ran in parallel, it will
# not be when ran in this mode)...
print("-- Running serially --")
e = engines.load(song, engine='serial')
e.run()
print("-- Statistics gathered --")
print(e.statistics)



#    Copyright (C) 2013 Yahoo! Inc. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import contextlib
import hashlib
import logging
import os
import random
import sys
import time

logging.basicConfig(level=logging.ERROR)

self_dir = os.path.abspath(os.path.dirname(__file__))
top_dir = os.path.abspath(os.path.join(os.path.dirname(__file__),
                                       os.pardir,
                                       os.pardir))
sys.path.insert(0, top_dir)
sys.path.insert(0, self_dir)

import futurist
from oslo_utils import uuidutils

from taskflow import engines
from taskflow import exceptions as exc
from taskflow.patterns import graph_flow as gf
from taskflow.patterns import linear_flow as lf
from taskflow.persistence import models
from taskflow import task

import example_utils as eu  # noqa

# INTRO: These examples show how a hierarchy of flows can be used to create a
# vm in a reliable & resumable manner using taskflow + a miniature version of
# what nova does while booting a vm.


@contextlib.contextmanager
def slow_down(how_long=0.5):
    try:
        yield how_long
    finally:
        if len(sys.argv) > 1:
            # Only both to do this if user input provided.
            print("** Ctrl-c me please!!! **")
            time.sleep(how_long)


class PrintText(task.Task):
    """Just inserts some text print outs in a workflow."""
    def __init__(self, print_what, no_slow=False):
        content_hash = hashlib.md5(print_what.encode('utf-8')).hexdigest()[0:8]
        super().__init__(name="Print: %s" % (content_hash))
        self._text = print_what
        self._no_slow = no_slow

    def execute(self):
        if self._no_slow:
            eu.print_wrapped(self._text)
        else:
            with slow_down():
                eu.print_wrapped(self._text)


class DefineVMSpec(task.Task):
    """Defines a vm specification to be."""
    def __init__(self, name):
        super().__init__(provides='vm_spec', name=name)

    def execute(self):
        return {
            'type': 'kvm',
            'disks': 2,
            'vcpu': 1,
            'ips': 1,
            'volumes': 3,
        }


class LocateImages(task.Task):
    """Locates where the vm images are."""
    def __init__(self, name):
        super().__init__(provides='image_locations', name=name)

    def execute(self, vm_spec):
        image_locations = {}
        for i in range(0, vm_spec['disks']):
            url = "http://www.yahoo.com/images/%s" % (i)
            image_locations[url] = "/tmp/%s.img" % (i)
        return image_locations


class DownloadImages(task.Task):
    """Downloads all the vm images."""
    def __init__(self, name):
        super().__init__(provides='download_paths',
                         name=name)

    def execute(self, image_locations):
        for src, loc in image_locations.items():
            with slow_down(1):
                print(f"Downloading from {src} => {loc}")
        return sorted(image_locations.values())


class CreateNetworkTpl(task.Task):
    """Generates the network settings file to be placed in the images."""
    SYSCONFIG_CONTENTS = """DEVICE=eth%s
BOOTPROTO=static
IPADDR=%s
ONBOOT=yes"""

    def __init__(self, name):
        super().__init__(provides='network_settings',
                         name=name)

    def execute(self, ips):
        settings = []
        for i, ip in enumerate(ips):
            settings.append(self.SYSCONFIG_CONTENTS % (i, ip))
        return settings


class AllocateIP(task.Task):
    """Allocates the ips for the given vm."""
    def __init__(self, name):
        super().__init__(provides='ips', name=name)

    def execute(self, vm_spec):
        ips = []
        for _i in range(0, vm_spec.get('ips', 0)):
            ips.append("192.168.0.%s" % (random.randint(1, 254)))
        return ips


class WriteNetworkSettings(task.Task):
    """Writes all the network settings into the downloaded images."""
    def execute(self, download_paths, network_settings):
        for j, path in enumerate(download_paths):
            with slow_down(1):
                print(f"Mounting {path} to /tmp/{j}")
            for i, setting in enumerate(network_settings):
                filename = ("/tmp/etc/sysconfig/network-scripts/"
                            "ifcfg-eth%s" % (i))
                with slow_down(1):
                    print("Writing to %s" % (filename))
                    print(setting)


class BootVM(task.Task):
    """Fires off the vm boot operation."""
    def execute(self, vm_spec):
        print("Starting vm!")
        with slow_down(1):
            print("Created: %s" % (vm_spec))


class AllocateVolumes(task.Task):
    """Allocates the volumes for the vm."""
    def execute(self, vm_spec):
        volumes = []
        for i in range(0, vm_spec['volumes']):
            with slow_down(1):
                volumes.append("/dev/vda%s" % (i + 1))
                print("Allocated volume %s" % volumes[-1])
        return volumes


class FormatVolumes(task.Task):
    """Formats the volumes for the vm."""
    def execute(self, volumes):
        for v in volumes:
            print("Formatting volume %s" % v)
            with slow_down(1):
                pass
            print("Formatted volume %s" % v)


def create_flow():
    # Setup the set of things to do (mini-nova).
    flow = lf.Flow("root").add(
        PrintText("Starting vm creation.", no_slow=True),
        lf.Flow('vm-maker').add(
            # First create a specification for the final vm to-be.
            DefineVMSpec("define_spec"),
            # This does all the image stuff.
            gf.Flow("img-maker").add(
                LocateImages("locate_images"),
                DownloadImages("download_images"),
            ),
            # This does all the network stuff.
            gf.Flow("net-maker").add(
                AllocateIP("get_my_ips"),
                CreateNetworkTpl("fetch_net_settings"),
                WriteNetworkSettings("write_net_settings"),
            ),
            # This does all the volume stuff.
            gf.Flow("volume-maker").add(
                AllocateVolumes("allocate_my_volumes", provides='volumes'),
                FormatVolumes("volume_formatter"),
            ),
            # Finally boot it all.
            BootVM("boot-it"),
        ),
        # Ya it worked!
        PrintText("Finished vm create.", no_slow=True),
        PrintText("Instance is running!", no_slow=True))
    return flow

eu.print_wrapped("Initializing")

# Setup the persistence & resumption layer.
with eu.get_backend() as backend:

    # Try to find a previously passed in tracking id...
    try:
        book_id, flow_id = sys.argv[2].split("+", 1)
        if not uuidutils.is_uuid_like(book_id):
            book_id = None
        if not uuidutils.is_uuid_like(flow_id):
            flow_id = None
    except (IndexError, ValueError):
        book_id = None
        flow_id = None

    # Set up how we want our engine to run, serial, parallel...
    try:
        executor = futurist.GreenThreadPoolExecutor(max_workers=5)
    except RuntimeError:
        # No eventlet installed, just let the default be used instead.
        executor = None

    # Create/fetch a logbook that will track the workflows work.
    book = None
    flow_detail = None
    if all([book_id, flow_id]):
        # Try to find in a prior logbook and flow detail...
        with contextlib.closing(backend.get_connection()) as conn:
            try:
                book = conn.get_logbook(book_id)
                flow_detail = book.find(flow_id)
            except exc.NotFound:
                pass
    if book is None and flow_detail is None:
        book = models.LogBook("vm-boot")
        with contextlib.closing(backend.get_connection()) as conn:
            conn.save_logbook(book)
        engine = engines.load_from_factory(create_flow,
                                           backend=backend, book=book,
                                           engine='parallel',
                                           executor=executor)
        print("!! Your tracking id is: '{}+{}'".format(
            book.uuid, engine.storage.flow_uuid))
        print("!! Please submit this on later runs for tracking purposes")
    else:
        # Attempt to load from a previously partially completed flow.
        engine = engines.load_from_detail(flow_detail, backend=backend,
                                          engine='parallel', executor=executor)

    # Make me my vm please!
    eu.print_wrapped('Running')
    engine.run()

# How to use.
#
# 1. $ python me.py "sqlite:////tmp/nova.db"
# 2. ctrl-c before this finishes
# 3. Find the tracking id (search for 'Your tracking id is')
# 4. $ python me.py "sqlite:////tmp/cinder.db" "$tracking_id"
# 5. Watch it pick up where it left off.
# 6. Profit!



#    Copyright (C) 2014 Yahoo! Inc. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging
import os
import sys

logging.basicConfig(level=logging.ERROR)

self_dir = os.path.abspath(os.path.dirname(__file__))
top_dir = os.path.abspath(os.path.join(os.path.dirname(__file__),
                                       os.pardir,
                                       os.pardir))
sys.path.insert(0, top_dir)
sys.path.insert(0, self_dir)

from taskflow import engines
from taskflow.patterns import linear_flow
from taskflow import task

# INTRO: This example shows how a task (in a linear/serial workflow) can
# produce an output that can be then consumed/used by a downstream task.


class TaskA(task.Task):
    default_provides = 'a'

    def execute(self):
        print("Executing '%s'" % (self.name))
        return 'a'


class TaskB(task.Task):
    def execute(self, a):
        print("Executing '%s'" % (self.name))
        print("Got input '%s'" % (a))


print("Constructing...")
wf = linear_flow.Flow("pass-from-to")
wf.add(TaskA('a'), TaskB('b'))

print("Loading...")
e = engines.load(wf)

print("Compiling...")
e.compile()

print("Preparing...")
e.prepare()

print("Running...")
e.run()

print("Done...")



#    Copyright (C) 2012-2013 Yahoo! Inc. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from taskflow.engines.action_engine.actions import base
from taskflow import retry as retry_atom
from taskflow import states
from taskflow.types import failure


class RetryAction(base.Action):
    """An action that handles executing, state changes, ... of retry atoms."""

    def __init__(self, storage, notifier, retry_executor):
        super().__init__(storage, notifier)
        self._retry_executor = retry_executor

    def _get_retry_args(self, retry, revert=False, addons=None):
        if revert:
            arguments = self._storage.fetch_mapped_args(
                retry.revert_rebind,
                atom_name=retry.name,
                optional_args=retry.revert_optional
            )
        else:
            arguments = self._storage.fetch_mapped_args(
                retry.rebind,
                atom_name=retry.name,
                optional_args=retry.optional
            )
        history = self._storage.get_retry_history(retry.name)
        arguments[retry_atom.EXECUTE_REVERT_HISTORY] = history
        if addons:
            arguments.update(addons)
        return arguments

    def change_state(self, retry, state, result=base.Action.NO_RESULT):
        old_state = self._storage.get_atom_state(retry.name)
        if state in self.SAVE_RESULT_STATES:
            save_result = None
            if result is not self.NO_RESULT:
                save_result = result
            self._storage.save(retry.name, save_result, state)
            # TODO(harlowja): combine this with the save to avoid a call
            # back into the persistence layer...
            if state == states.REVERTED:
                self._storage.cleanup_retry_history(retry.name, state)
        else:
            if state == old_state:
                # NOTE(imelnikov): nothing really changed, so we should not
                # write anything to storage and run notifications.
                return
            self._storage.set_atom_state(retry.name, state)
        retry_uuid = self._storage.get_atom_uuid(retry.name)
        details = {
            'retry_name': retry.name,
            'retry_uuid': retry_uuid,
            'old_state': old_state,
        }
        if result is not self.NO_RESULT:
            details['result'] = result
        self._notifier.notify(state, details)

    def schedule_execution(self, retry):
        self.change_state(retry, states.RUNNING)
        return self._retry_executor.execute_retry(
            retry, self._get_retry_args(retry))

    def complete_reversion(self, retry, result):
        if isinstance(result, failure.Failure):
            self.change_state(retry, states.REVERT_FAILURE, result=result)
        else:
            self.change_state(retry, states.REVERTED, result=result)

    def complete_execution(self, retry, result):
        if isinstance(result, failure.Failure):
            self.change_state(retry, states.FAILURE, result=result)
        else:
            self.change_state(retry, states.SUCCESS, result=result)

    def schedule_reversion(self, retry):
        self.change_state(retry, states.REVERTING)
        arg_addons = {
            retry_atom.REVERT_FLOW_FAILURES: self._storage.get_failures(),
        }
        return self._retry_executor.revert_retry(
            retry, self._get_retry_args(retry, addons=arg_addons, revert=True))

    def on_failure(self, retry, atom, last_failure):
        self._storage.save_retry_failure(retry.name, atom.name, last_failure)
        arguments = self._get_retry_args(retry)
        return retry.on_failure(**arguments)



#    Copyright (C) 2012 Yahoo! Inc. All Rights Reserved.
#    Copyright (C) 2013 Rackspace Hosting All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import contextlib
import errno
import os
import shutil

import cachetools
import fasteners
from oslo_serialization import jsonutils
from oslo_utils import fileutils

from taskflow import exceptions as exc
from taskflow.persistence import path_based
from taskflow.utils import misc


@contextlib.contextmanager
def _storagefailure_wrapper():
    try:
        yield
    except exc.TaskFlowException:
        raise
    except Exception as e:
        if isinstance(e, (IOError, OSError)) and e.errno == errno.ENOENT:
            exc.raise_with_cause(exc.NotFound,
                                 'Item not found: %s' % e.filename,
                                 cause=e)
        else:
            exc.raise_with_cause(exc.StorageFailure,
                                 "Storage backend internal error", cause=e)


class DirBackend(path_based.PathBasedBackend):
    """A directory and file based backend.

    This backend does *not* provide true transactional semantics. It does
    guarantee that there will be no interprocess race conditions when
    writing and reading by using a consistent hierarchy of file based locks.

    Example configuration::

        conf = {
            "path": "/tmp/taskflow",  # save data to this root directory
            "max_cache_size": 1024,  # keep up-to 1024 entries in memory
        }
    """

    DEFAULT_FILE_ENCODING = 'utf-8'
    """
    Default encoding used when decoding or encoding files into or from
    text/unicode into binary or binary into text/unicode.
    """

    def __init__(self, conf):
        super().__init__(conf)
        max_cache_size = self._conf.get('max_cache_size')
        if max_cache_size is not None:
            max_cache_size = int(max_cache_size)
            if max_cache_size < 1:
                raise ValueError("Maximum cache size must be greater than"
                                 " or equal to one")
            self.file_cache = cachetools.LRUCache(max_cache_size)
        else:
            self.file_cache = {}
        self.encoding = self._conf.get('encoding', self.DEFAULT_FILE_ENCODING)
        if not self._path:
            raise ValueError("Empty path is disallowed")
        self._path = os.path.abspath(self._path)
        self.lock = fasteners.ReaderWriterLock()

    def get_connection(self):
        return Connection(self)

    def close(self):
        pass


class Connection(path_based.PathBasedConnection):
    def _read_from(self, filename):
        # This is very similar to the oslo-incubator fileutils module, but
        # tweaked to not depend on a global cache, as well as tweaked to not
        # pull-in the oslo logging module (which is a huge pile of code).
        mtime = os.path.getmtime(filename)
        cache_info = self.backend.file_cache.setdefault(filename, {})
        if not cache_info or mtime > cache_info.get('mtime', 0):
            with open(filename, encoding=self.backend.encoding) as fp:
                cache_info['data'] = fp.read()
                cache_info['mtime'] = mtime
        return cache_info['data']

    def _write_to(self, filename, contents):
        contents = misc.binary_encode(contents,
                                      encoding=self.backend.encoding)
        with open(filename, 'wb') as fp:
            fp.write(contents)
        self.backend.file_cache.pop(filename, None)

    @contextlib.contextmanager
    def _path_lock(self, path):
        lockfile = self._join_path(path, 'lock')
        with fasteners.InterProcessLock(lockfile) as lock:
            with _storagefailure_wrapper():
                yield lock

    def _join_path(self, *parts):
        return os.path.join(*parts)

    def _get_item(self, path):
        with self._path_lock(path):
            item_path = self._join_path(path, 'metadata')
            return misc.decode_json(self._read_from(item_path))

    def _set_item(self, path, value, transaction):
        with self._path_lock(path):
            item_path = self._join_path(path, 'metadata')
            self._write_to(item_path, jsonutils.dumps(value))

    def _del_tree(self, path, transaction):
        with self._path_lock(path):
            shutil.rmtree(path)

    def _get_children(self, path):
        if path == self.book_path:
            filter_func = os.path.isdir
        else:
            filter_func = os.path.islink
        with _storagefailure_wrapper():
            return [child for child in os.listdir(path)
                    if filter_func(self._join_path(path, child))]

    def _ensure_path(self, path):
        with _storagefailure_wrapper():
            fileutils.ensure_tree(path)

    def _create_link(self, src_path, dest_path, transaction):
        with _storagefailure_wrapper():
            try:
                os.symlink(src_path, dest_path)
            except OSError as e:
                if e.errno != errno.EEXIST:
                    raise

    @contextlib.contextmanager
    def _transaction(self):
        """This just wraps a global write-lock."""
        lock = self.backend.lock.write_lock
        with lock():
            yield

    def validate(self):
        with _storagefailure_wrapper():
            for p in (self.flow_path, self.atom_path, self.book_path):
                if not os.path.isdir(p):
                    raise RuntimeError("Missing required directory: %s" % (p))



#    Copyright (C) 2015 Rackspace Hosting All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc

from taskflow import exceptions as exc
from taskflow.persistence import base
from taskflow.persistence import models


class PathBasedBackend(base.Backend, metaclass=abc.ABCMeta):
    """Base class for persistence backends that address data by path

    Subclasses of this backend write logbooks, flow details, and atom details
    to a provided base path in some filesystem-like storage. They will create
    and store those objects in three key directories (one for logbooks, one
    for flow details and one for atom details). They create those associated
    directories and then create files inside those directories that represent
    the contents of those objects for later reading and writing.
    """

    #: Default path used when none is provided.
    DEFAULT_PATH = None

    def __init__(self, conf):
        super().__init__(conf)
        self._path = self._conf.get('path', None)
        if not self._path:
            self._path = self.DEFAULT_PATH

    @property
    def path(self):
        return self._path


class PathBasedConnection(base.Connection, metaclass=abc.ABCMeta):
    """Base class for path based backend connections."""

    def __init__(self, backend):
        self._backend = backend
        self._book_path = self._join_path(backend.path, "books")
        self._flow_path = self._join_path(backend.path, "flow_details")
        self._atom_path = self._join_path(backend.path, "atom_details")

    @staticmethod
    def _serialize(obj):
        if isinstance(obj, models.LogBook):
            return obj.to_dict(marshal_time=True)
        elif isinstance(obj, models.FlowDetail):
            return obj.to_dict()
        elif isinstance(obj, models.AtomDetail):
            return base._format_atom(obj)
        else:
            raise exc.StorageFailure("Invalid storage class %s" % type(obj))

    @staticmethod
    def _deserialize(cls, data):
        if issubclass(cls, models.LogBook):
            return cls.from_dict(data, unmarshal_time=True)
        elif issubclass(cls, models.FlowDetail):
            return cls.from_dict(data)
        elif issubclass(cls, models.AtomDetail):
            atom_class = models.atom_detail_class(data['type'])
            return atom_class.from_dict(data['atom'])
        else:
            raise exc.StorageFailure("Invalid storage class %s" % cls)

    @property
    def backend(self):
        return self._backend

    @property
    def book_path(self):
        return self._book_path

    @property
    def flow_path(self):
        return self._flow_path

    @property
    def atom_path(self):
        return self._atom_path

    @abc.abstractmethod
    def _join_path(self, *parts):
        """Accept path parts, and return a joined path"""

    @abc.abstractmethod
    def _get_item(self, path):
        """Fetch a single item from the backend"""

    @abc.abstractmethod
    def _set_item(self, path, value, transaction):
        """Write a single item to the backend"""

    @abc.abstractmethod
    def _del_tree(self, path, transaction):
        """Recursively deletes a folder from the backend."""

    @abc.abstractmethod
    def _get_children(self, path):
        """Get a list of child items of a path"""

    @abc.abstractmethod
    def _ensure_path(self, path):
        """Recursively ensure that a path (folder) in the backend exists"""

    @abc.abstractmethod
    def _create_link(self, src_path, dest_path, transaction):
        """Create a symlink-like link between two paths"""

    @abc.abstractmethod
    def _transaction(self):
        """Context manager that yields a transaction"""

    def _get_obj_path(self, obj):
        if isinstance(obj, models.LogBook):
            path = self.book_path
        elif isinstance(obj, models.FlowDetail):
            path = self.flow_path
        elif isinstance(obj, models.AtomDetail):
            path = self.atom_path
        else:
            raise exc.StorageFailure("Invalid storage class %s" % type(obj))
        return self._join_path(path, obj.uuid)

    def _update_object(self, obj, transaction, ignore_missing=False):
        path = self._get_obj_path(obj)
        try:
            item_data = self._get_item(path)
            existing_obj = self._deserialize(type(obj), item_data)
            obj = existing_obj.merge(obj)
        except exc.NotFound:
            if not ignore_missing:
                raise
        self._set_item(path, self._serialize(obj), transaction)
        return obj

    def get_logbooks(self, lazy=False):
        for book_uuid in self._get_children(self.book_path):
            yield self.get_logbook(book_uuid, lazy=lazy)

    def get_logbook(self, book_uuid, lazy=False):
        book_path = self._join_path(self.book_path, book_uuid)
        book_data = self._get_item(book_path)
        book = self._deserialize(models.LogBook, book_data)
        if not lazy:
            for flow_details in self.get_flows_for_book(book_uuid):
                book.add(flow_details)
        return book

    def save_logbook(self, book):
        book_path = self._get_obj_path(book)
        with self._transaction() as transaction:
            self._update_object(book, transaction, ignore_missing=True)
            for flow_details in book:
                flow_path = self._get_obj_path(flow_details)
                link_path = self._join_path(book_path, flow_details.uuid)
                self._do_update_flow_details(flow_details, transaction,
                                             ignore_missing=True)
                self._create_link(flow_path, link_path, transaction)
        return book

    def get_flows_for_book(self, book_uuid, lazy=False):
        book_path = self._join_path(self.book_path, book_uuid)
        for flow_uuid in self._get_children(book_path):
            yield self.get_flow_details(flow_uuid, lazy)

    def get_flow_details(self, flow_uuid, lazy=False):
        flow_path = self._join_path(self.flow_path, flow_uuid)
        flow_data = self._get_item(flow_path)
        flow_details = self._deserialize(models.FlowDetail, flow_data)
        if not lazy:
            for atom_details in self.get_atoms_for_flow(flow_uuid):
                flow_details.add(atom_details)
        return flow_details

    def _do_update_flow_details(self, flow_detail, transaction,
                                ignore_missing=False):
        flow_path = self._get_obj_path(flow_detail)
        self._update_object(flow_detail, transaction,
                            ignore_missing=ignore_missing)
        for atom_details in flow_detail:
            atom_path = self._get_obj_path(atom_details)
            link_path = self._join_path(flow_path, atom_details.uuid)
            self._create_link(atom_path, link_path, transaction)
            self._update_object(atom_details, transaction, ignore_missing=True)
        return flow_detail

    def update_flow_details(self, flow_detail, ignore_missing=False):
        with self._transaction() as transaction:
            return self._do_update_flow_details(flow_detail, transaction,
                                                ignore_missing=ignore_missing)

    def get_atoms_for_flow(self, flow_uuid):
        flow_path = self._join_path(self.flow_path, flow_uuid)
        for atom_uuid in self._get_children(flow_path):
            yield self.get_atom_details(atom_uuid)

    def get_atom_details(self, atom_uuid):
        atom_path = self._join_path(self.atom_path, atom_uuid)
        atom_data = self._get_item(atom_path)
        return self._deserialize(models.AtomDetail, atom_data)

    def update_atom_details(self, atom_detail, ignore_missing=False):
        with self._transaction() as transaction:
            return self._update_object(atom_detail, transaction,
                                       ignore_missing=ignore_missing)

    def _do_destroy_logbook(self, book_uuid, transaction):
        book_path = self._join_path(self.book_path, book_uuid)
        for flow_uuid in self._get_children(book_path):
            flow_path = self._join_path(self.flow_path, flow_uuid)
            for atom_uuid in self._get_children(flow_path):
                atom_path = self._join_path(self.atom_path, atom_uuid)
                self._del_tree(atom_path, transaction)
            self._del_tree(flow_path, transaction)
        self._del_tree(book_path, transaction)

    def destroy_logbook(self, book_uuid):
        with self._transaction() as transaction:
            return self._do_destroy_logbook(book_uuid, transaction)

    def clear_all(self):
        with self._transaction() as transaction:
            for path in (self.book_path, self.flow_path, self.atom_path):
                self._del_tree(path, transaction)

    def upgrade(self):
        for path in (self.book_path, self.flow_path, self.atom_path):
            self._ensure_path(path)

    def close(self):
        pass



#    Copyright (C) 2014 Yahoo! Inc. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections
import io
import itertools
import os

from taskflow.types import graph
from taskflow.utils import iter_utils
from taskflow.utils import misc


class FrozenNode(Exception):
    """Exception raised when a frozen node is modified."""

    def __init__(self):
        super().__init__("Frozen node(s) can't be modified")


class _DFSIter:
    """Depth first iterator (non-recursive) over the child nodes."""

    def __init__(self, root, include_self=False, right_to_left=True):
        self.root = root
        self.right_to_left = bool(right_to_left)
        self.include_self = bool(include_self)

    def __iter__(self):
        stack = []
        if self.include_self:
            stack.append(self.root)
        else:
            if self.right_to_left:
                stack.extend(self.root.reverse_iter())
            else:
                # Traverse the left nodes first to the right nodes.
                stack.extend(iter(self.root))
        while stack:
            # Visit the node.
            node = stack.pop()
            yield node
            if self.right_to_left:
                stack.extend(node.reverse_iter())
            else:
                # Traverse the left nodes first to the right nodes.
                stack.extend(iter(node))


class _BFSIter:
    """Breadth first iterator (non-recursive) over the child nodes."""

    def __init__(self, root, include_self=False, right_to_left=False):
        self.root = root
        self.right_to_left = bool(right_to_left)
        self.include_self = bool(include_self)

    def __iter__(self):
        q = collections.deque()
        if self.include_self:
            q.append(self.root)
        else:
            if self.right_to_left:
                q.extend(iter(self.root))
            else:
                # Traverse the left nodes first to the right nodes.
                q.extend(self.root.reverse_iter())
        while q:
            # Visit the node.
            node = q.popleft()
            yield node
            if self.right_to_left:
                q.extend(iter(node))
            else:
                # Traverse the left nodes first to the right nodes.
                q.extend(node.reverse_iter())


class Node:
    """A n-ary node class that can be used to create tree structures."""

    #: Default string prefix used in :py:meth:`.pformat`.
    STARTING_PREFIX = ""

    #: Default string used to create empty space used in :py:meth:`.pformat`.
    EMPTY_SPACE_SEP = " "

    HORIZONTAL_CONN = "__"
    """
    Default string used to horizontally connect a node to its
    parent (used in :py:meth:`.pformat`.).
    """

    VERTICAL_CONN = "|"
    """
    Default string used to vertically connect a node to its
    parent (used in :py:meth:`.pformat`).
    """

    #: Default line separator used in :py:meth:`.pformat`.
    LINE_SEP = os.linesep

    def __init__(self, item, **kwargs):
        self.item = item
        self.parent = None
        self.metadata = dict(kwargs)
        self.frozen = False
        self._children = []

    def freeze(self):
        if not self.frozen:
            # This will DFS until all children are frozen as well, only
            # after that works do we freeze ourselves (this makes it so
            # that we don't become frozen if a child node fails to perform
            # the freeze operation).
            for n in self:
                n.freeze()
            self.frozen = True

    @misc.disallow_when_frozen(FrozenNode)
    def add(self, child):
        """Adds a child to this node (appends to left of existing children).

        NOTE(harlowja): this will also set the childs parent to be this node.
        """
        child.parent = self
        self._children.append(child)

    def empty(self):
        """Returns if the node is a leaf node."""
        return self.child_count() == 0

    def path_iter(self, include_self=True):
        """Yields back the path from this node to the root node."""
        if include_self:
            node = self
        else:
            node = self.parent
        while node is not None:
            yield node
            node = node.parent

    def find_first_match(self, matcher, only_direct=False, include_self=True):
        """Finds the *first* node that matching callback returns true.

        This will search not only this node but also any children nodes (in
        depth first order, from right to left) and finally if nothing is
        matched then ``None`` is returned instead of a node object.

        :param matcher: callback that takes one positional argument (a node)
                        and returns true if it matches desired node or false
                        if not.
        :param only_direct: only look at current node and its
                            direct children (implies that this does not
                            search using depth first).
        :param include_self: include the current node during searching.

        :returns: the node that matched (or ``None``)
        """
        if only_direct:
            if include_self:
                it = itertools.chain([self], self.reverse_iter())
            else:
                it = self.reverse_iter()
        else:
            it = self.dfs_iter(include_self=include_self)
        return iter_utils.find_first_match(it, matcher)

    def find(self, item, only_direct=False, include_self=True):
        """Returns the *first* node for an item if it exists in this node.

        This will search not only this node but also any children nodes (in
        depth first order, from right to left) and finally if nothing is
        matched then ``None`` is returned instead of a node object.

        :param item: item to look for.
        :param only_direct: only look at current node and its
                            direct children (implies that this does not
                            search using depth first).
        :param include_self: include the current node during searching.

        :returns: the node that matched provided item (or ``None``)
        """
        return self.find_first_match(lambda n: n.item == item,
                                     only_direct=only_direct,
                                     include_self=include_self)

    @misc.disallow_when_frozen(FrozenNode)
    def disassociate(self):
        """Removes this node from its parent (if any).

        :returns: occurrences of this node that were removed from its parent.
        """
        occurrences = 0
        if self.parent is not None:
            p = self.parent
            self.parent = None
            # Remove all instances of this node from its parent.
            while True:
                try:
                    p._children.remove(self)
                except ValueError:
                    break
                else:
                    occurrences += 1
        return occurrences

    @misc.disallow_when_frozen(FrozenNode)
    def remove(self, item, only_direct=False, include_self=True):
        """Removes a item from this nodes children.

        This will search not only this node but also any children nodes and
        finally if nothing is found then a value error is raised instead of
        the normally returned *removed* node object.

        :param item: item to lookup.
        :param only_direct: only look at current node and its
                            direct children (implies that this does not
                            search using depth first).
        :param include_self: include the current node during searching.
        """
        node = self.find(item, only_direct=only_direct,
                         include_self=include_self)
        if node is None:
            raise ValueError("Item '%s' not found to remove" % item)
        else:
            node.disassociate()
            return node

    def __contains__(self, item):
        """Returns whether item exists in this node or this nodes children.

        :returns: if the item exists in this node or nodes children,
                  true if the item exists, false otherwise
        :rtype: boolean
        """
        return self.find(item) is not None

    def __getitem__(self, index):
        # NOTE(harlowja): 0 is the right most index, len - 1 is the left most
        return self._children[index]

    def pformat(self, stringify_node=None,
                linesep=LINE_SEP, vertical_conn=VERTICAL_CONN,
                horizontal_conn=HORIZONTAL_CONN, empty_space=EMPTY_SPACE_SEP,
                starting_prefix=STARTING_PREFIX):
        """Formats this node + children into a nice string representation.

        **Example**::

            >>> from taskflow.types import tree
            >>> yahoo = tree.Node("CEO")
            >>> yahoo.add(tree.Node("Infra"))
            >>> yahoo[0].add(tree.Node("Boss"))
            >>> yahoo[0][0].add(tree.Node("Me"))
            >>> yahoo.add(tree.Node("Mobile"))
            >>> yahoo.add(tree.Node("Mail"))
            >>> print(yahoo.pformat())
            CEO
            |__Infra
            |  |__Boss
            |     |__Me
            |__Mobile
            |__Mail
        """
        if stringify_node is None:
            # Default to making a unicode string out of the nodes item...
            stringify_node = lambda node: str(node.item)
        expected_lines = self.child_count(only_direct=False) + 1
        buff = io.StringIO()
        conn = vertical_conn + horizontal_conn
        stop_at_parent = self
        for i, node in enumerate(self.dfs_iter(include_self=True), 1):
            prefix = []
            connected_to_parent = False
            last_node = node
            # Walk through *most* of this nodes parents, and form the expected
            # prefix that each parent should require, repeat this until we
            # hit the root node (self) and use that as our nodes prefix
            # string...
            parent_node_it = iter_utils.while_is_not(
                node.path_iter(include_self=True), stop_at_parent)
            for j, parent_node in enumerate(parent_node_it):
                if parent_node is stop_at_parent:
                    if j > 0:
                        if not connected_to_parent:
                            prefix.append(conn)
                            connected_to_parent = True
                        else:
                            # If the node was connected already then it must
                            # have had more than one parent, so we want to put
                            # the right final starting prefix on (which may be
                            # a empty space or another vertical connector)...
                            last_node = self._children[-1]
                            m = last_node.find_first_match(lambda n: n is node,
                                                           include_self=False,
                                                           only_direct=False)
                            if m is not None:
                                prefix.append(empty_space)
                            else:
                                prefix.append(vertical_conn)
                elif parent_node is node:
                    # Skip ourself... (we only include ourself so that
                    # we can use the 'j' variable to determine if the only
                    # node requested is ourself in the first place); used
                    # in the first conditional here...
                    pass
                else:
                    if not connected_to_parent:
                        prefix.append(conn)
                        spaces = len(horizontal_conn)
                        connected_to_parent = True
                    else:
                        # If we have already been connected to our parent
                        # then determine if this current node is the last
                        # node of its parent (and in that case just put
                        # on more spaces), otherwise put a vertical connector
                        # on and less spaces...
                        if parent_node[-1] is not last_node:
                            prefix.append(vertical_conn)
                            spaces = len(horizontal_conn)
                        else:
                            spaces = len(conn)
                    prefix.append(empty_space * spaces)
                last_node = parent_node
            prefix.append(starting_prefix)
            for prefix_piece in reversed(prefix):
                buff.write(prefix_piece)
            buff.write(stringify_node(node))
            if i != expected_lines:
                buff.write(linesep)
        return buff.getvalue()

    def child_count(self, only_direct=True):
        """Returns how many children this node has.

        This can be either only the direct children of this node or inclusive
        of all children nodes of this node (children of children and so-on).

        NOTE(harlowja): it does not account for the current node in this count.
        """
        if not only_direct:
            return iter_utils.count(self.dfs_iter())
        return len(self._children)

    def __iter__(self):
        """Iterates over the direct children of this node (right->left)."""
        yield from self._children

    def reverse_iter(self):
        """Iterates over the direct children of this node (left->right)."""
        yield from reversed(self._children)

    def index(self, item):
        """Finds the child index of a given item, searches in added order."""
        index_at = None
        for (i, child) in enumerate(self._children):
            if child.item == item:
                index_at = i
                break
        if index_at is None:
            raise ValueError("%s is not contained in any child" % (item))
        return index_at

    def dfs_iter(self, include_self=False, right_to_left=True):
        """Depth first iteration (non-recursive) over the child nodes."""
        return _DFSIter(self,
                        include_self=include_self,
                        right_to_left=right_to_left)

    def bfs_iter(self, include_self=False, right_to_left=False):
        """Breadth first iteration (non-recursive) over the child nodes."""
        return _BFSIter(self,
                        include_self=include_self,
                        right_to_left=right_to_left)

    def to_digraph(self):
        """Converts this node + its children into a ordered directed graph.

        The graph returned will have the same structure as the
        this node and its children (and tree node metadata will be translated
        into graph node metadata).

        :returns: a directed graph
        :rtype: :py:class:`taskflow.types.graph.OrderedDiGraph`
        """
        g = graph.OrderedDiGraph()
        for node in self.bfs_iter(include_self=True, right_to_left=True):
            g.add_node(node.item, **node.metadata)
            if node is not self:
                g.add_edge(node.parent.item, node.item)
        return g



#    Copyright (C) 2014 Yahoo! Inc. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import threading

import fasteners
from oslo_utils import excutils

from taskflow import flow
from taskflow import logging
from taskflow import task
from taskflow.types import graph as gr
from taskflow.types import tree as tr
from taskflow.utils import iter_utils
from taskflow.utils import misc

from taskflow.flow import (LINK_INVARIANT, LINK_RETRY)  # noqa

LOG = logging.getLogger(__name__)

# Constants attached to node attributes in the execution graph (and tree
# node metadata), provided as constants here and constants in the compilation
# class (so that users will not have to import this file to access them); but
# provide them as module constants so that internal code can more
# easily access them...
TASK = 'task'
RETRY = 'retry'
FLOW = 'flow'
FLOW_END = 'flow_end'

# Quite often used together, so make a tuple everyone can share...
ATOMS = (TASK, RETRY)
FLOWS = (FLOW, FLOW_END)


class Terminator:
    """Flow terminator class."""

    def __init__(self, flow):
        self._flow = flow
        self._name = f"{self._flow.name}[$]"

    @property
    def flow(self):
        """The flow which this terminator signifies/marks the end of."""
        return self._flow

    @property
    def name(self):
        """Useful name this end terminator has (derived from flow name)."""
        return self._name

    def __str__(self):
        flow_name = ("%s" % self._flow).strip('"')
        return '"%s[$]"' % flow_name


class Compilation:
    """The result of a compilers ``compile()`` is this *immutable* object."""

    #: Task nodes will have a ``kind`` metadata key with this value.
    TASK = TASK

    #: Retry nodes will have a ``kind`` metadata key with this value.
    RETRY = RETRY

    FLOW = FLOW
    """
    Flow **entry** nodes will have a ``kind`` metadata key with
    this value.
    """

    FLOW_END = FLOW_END
    """
    Flow **exit** nodes will have a ``kind`` metadata key with
    this value (only applicable for compilation execution graph, not currently
    used in tree hierarchy).
    """

    def __init__(self, execution_graph, hierarchy):
        self._execution_graph = execution_graph
        self._hierarchy = hierarchy

    @property
    def execution_graph(self):
        """The execution ordering of atoms (as a graph structure)."""
        return self._execution_graph

    @property
    def hierarchy(self):
        """The hierarchy of patterns (as a tree structure)."""
        return self._hierarchy


def _overlap_occurrence_detector(to_graph, from_graph):
    """Returns how many nodes in 'from' graph are in 'to' graph (if any)."""
    return iter_utils.count(node for node in from_graph.nodes
                            if node in to_graph)


def _add_update_edges(graph, nodes_from, nodes_to, attr_dict=None):
    """Adds/updates edges from nodes to other nodes in the specified graph.

    It will connect the 'nodes_from' to the 'nodes_to' if an edge currently
    does *not* exist (if it does already exist then the edges attributes
    are just updated instead). When an edge is created the provided edge
    attributes dictionary will be applied to the new edge between these two
    nodes.
    """
    # NOTE(harlowja): give each edge its own attr copy so that if it's
    # later modified that the same copy isn't modified...
    for u in nodes_from:
        for v in nodes_to:
            if not graph.has_edge(u, v):
                if attr_dict:
                    graph.add_edge(u, v, attr_dict=attr_dict.copy())
                else:
                    graph.add_edge(u, v)
            else:
                # Just update the attr_dict (if any).
                if attr_dict:
                    graph.add_edge(u, v, attr_dict=attr_dict.copy())


class TaskCompiler:
    """Non-recursive compiler of tasks."""

    def compile(self, task, parent=None):
        graph = gr.DiGraph(name=task.name)
        graph.add_node(task, kind=TASK)
        node = tr.Node(task, kind=TASK)
        if parent is not None:
            parent.add(node)
        return graph, node


class FlowCompiler:
    """Recursive compiler of flows."""

    def __init__(self, deep_compiler_func):
        self._deep_compiler_func = deep_compiler_func

    def compile(self, flow, parent=None):
        """Decomposes a flow into a graph and scope tree hierarchy."""
        graph = gr.DiGraph(name=flow.name)
        graph.add_node(flow, kind=FLOW, noop=True)
        tree_node = tr.Node(flow, kind=FLOW, noop=True)
        if parent is not None:
            parent.add(tree_node)
        if flow.retry is not None:
            tree_node.add(tr.Node(flow.retry, kind=RETRY))
        decomposed = {
            child: self._deep_compiler_func(child, parent=tree_node)[0]
            for child in flow}
        decomposed_graphs = list(decomposed.values())
        graph = gr.merge_graphs(graph, *decomposed_graphs,
                                overlap_detector=_overlap_occurrence_detector)
        for u, v, attr_dict in flow.iter_links():
            u_graph = decomposed[u]
            v_graph = decomposed[v]
            _add_update_edges(graph, u_graph.no_successors_iter(),
                              list(v_graph.no_predecessors_iter()),
                              attr_dict=attr_dict)
        # Insert the flow(s) retry if needed, and always make sure it
        # is the **immediate** successor of the flow node itself.
        if flow.retry is not None:
            graph.add_node(flow.retry, kind=RETRY)
            _add_update_edges(graph, [flow], [flow.retry],
                              attr_dict={LINK_INVARIANT: True})
            for node in graph.nodes:
                if node is not flow.retry and node is not flow:
                    graph.nodes[node].setdefault(RETRY, flow.retry)
            from_nodes = [flow.retry]
            attr_dict = {LINK_INVARIANT: True, LINK_RETRY: True}
        else:
            from_nodes = [flow]
            attr_dict = {LINK_INVARIANT: True}
        # Ensure all nodes with no predecessors are connected to this flow
        # or its retry node (so that the invariant that the flow node is
        # traversed through before its contents is maintained); this allows
        # us to easily know when we have entered a flow (when running) and
        # do special and/or smart things such as only traverse up to the
        # start of a flow when looking for node deciders.
        _add_update_edges(graph, from_nodes, [
            node for node in graph.no_predecessors_iter()
            if node is not flow
        ], attr_dict=attr_dict)
        # Connect all nodes with no successors into a special terminator
        # that is used to identify the end of the flow and ensure that all
        # execution traversals will traverse over this node before executing
        # further work (this is especially useful for nesting and knowing
        # when we have exited a nesting level); it allows us to do special
        # and/or smart things such as applying deciders up to (but not
        # beyond) a flow termination point.
        #
        # Do note that in a empty flow this will just connect itself to
        # the flow node itself... and also note we can not use the flow
        # object itself (primarily because the underlying graph library
        # uses hashing to identify node uniqueness and we can easily create
        # a loop if we don't do this correctly, so avoid that by just
        # creating this special node and tagging it with a special kind); we
        # may be able to make this better in the future with a multidigraph
        # that networkx provides??
        flow_term = Terminator(flow)
        graph.add_node(flow_term, kind=FLOW_END, noop=True)
        _add_update_edges(graph, [
            node for node in graph.no_successors_iter()
            if node is not flow_term
        ], [flow_term], attr_dict={LINK_INVARIANT: True})
        return graph, tree_node


class PatternCompiler:
    """Compiles a flow pattern (or task) into a compilation unit.

    Let's dive into the basic idea for how this works:

    The compiler here is provided a 'root' object via its __init__ method,
    this object could be a task, or a flow (one of the supported patterns),
    the end-goal is to produce a :py:class:`.Compilation` object as the result
    with the needed components. If this is not possible a
    :py:class:`~.taskflow.exceptions.CompilationFailure` will be raised.
    In the case where a **unknown** type is being requested to compile
    a ``TypeError`` will be raised and when a duplicate object (one that
    has **already** been compiled) is encountered a ``ValueError`` is raised.

    The complexity of this comes into play when the 'root' is a flow that
    contains itself other nested flows (and so-on); to compile this object and
    its contained objects into a graph that *preserves* the constraints the
    pattern mandates we have to go through a recursive algorithm that creates
    subgraphs for each nesting level, and then on the way back up through
    the recursion (now with a decomposed mapping from contained patterns or
    atoms to there corresponding subgraph) we have to then connect the
    subgraphs (and the atom(s) there-in) that were decomposed for a pattern
    correctly into a new graph and then ensure the pattern mandated
    constraints are retained. Finally we then return to the
    caller (and they will do the same thing up until the root node, which by
    that point one graph is created with all contained atoms in the
    pattern/nested patterns mandated ordering).

    Also maintained in the :py:class:`.Compilation` object is a hierarchy of
    the nesting of items (which is also built up during the above mentioned
    recusion, via a much simpler algorithm); this is typically used later to
    determine the prior atoms of a given atom when looking up values that can
    be provided to that atom for execution (see the scopes.py file for how this
    works). Note that although you *could* think that the graph itself could be
    used for this, which in some ways it can (for limited usage) the hierarchy
    retains the nested structure (which is useful for scoping analysis/lookup)
    to be able to provide back a iterator that gives back the scopes visible
    at each level (the graph does not have this information once flattened).

    Let's take an example:

    Given the pattern ``f(a(b, c), d)`` where ``f`` is a
    :py:class:`~taskflow.patterns.linear_flow.Flow` with items ``a(b, c)``
    where ``a`` is a :py:class:`~taskflow.patterns.linear_flow.Flow` composed
    of tasks ``(b, c)`` and task ``d``.

    The algorithm that will be performed (mirroring the above described logic)
    will go through the following steps (the tree hierarchy building is left
    out as that is more obvious)::

        Compiling f
          - Decomposing flow f with no parent (must be the root)
          - Compiling a
              - Decomposing flow a with parent f
              - Compiling b
                  - Decomposing task b with parent a
                  - Decomposed b into:
                    Name: b
                    Nodes: 1
                      - b
                    Edges: 0
              - Compiling c
                  - Decomposing task c with parent a
                  - Decomposed c into:
                    Name: c
                    Nodes: 1
                      - c
                    Edges: 0
              - Relinking decomposed b -> decomposed c
              - Decomposed a into:
                Name: a
                Nodes: 2
                  - b
                  - c
                Edges: 1
                  b -> c ({'invariant': True})
          - Compiling d
              - Decomposing task d with parent f
              - Decomposed d into:
                Name: d
                Nodes: 1
                  - d
                Edges: 0
          - Relinking decomposed a -> decomposed d
          - Decomposed f into:
            Name: f
            Nodes: 3
              - c
              - b
              - d
            Edges: 2
              c -> d ({'invariant': True})
              b -> c ({'invariant': True})
    """

    def __init__(self, root, freeze=True):
        self._root = root
        self._history = set()
        self._freeze = freeze
        self._lock = threading.Lock()
        self._compilation = None
        self._matchers = [
            (flow.Flow, FlowCompiler(self._compile)),
            (task.Task, TaskCompiler()),
        ]
        self._level = 0

    def _compile(self, item, parent=None):
        """Compiles a item (pattern, task) into a graph + tree node."""
        item_compiler = misc.match_type(item, self._matchers)
        if item_compiler is not None:
            self._pre_item_compile(item)
            graph, node = item_compiler.compile(item, parent=parent)
            self._post_item_compile(item, graph, node)
            return graph, node
        else:
            raise TypeError("Unknown object '%s' (%s) requested to compile"
                            % (item, type(item)))

    def _pre_item_compile(self, item):
        """Called before a item is compiled; any pre-compilation actions."""
        if item in self._history:
            raise ValueError("Already compiled item '%s' (%s), duplicate"
                             " and/or recursive compiling is not"
                             " supported" % (item, type(item)))
        self._history.add(item)
        if LOG.isEnabledFor(logging.TRACE):
            LOG.trace("%sCompiling '%s'", "  " * self._level, item)
        self._level += 1

    def _post_item_compile(self, item, graph, node):
        """Called after a item is compiled; doing post-compilation actions."""
        self._level -= 1
        if LOG.isEnabledFor(logging.TRACE):
            prefix = '  ' * self._level
            LOG.trace("%sDecomposed '%s' into:", prefix, item)
            prefix = '  ' * (self._level + 1)
            LOG.trace("%sGraph:", prefix)
            for line in graph.pformat().splitlines():
                LOG.trace("%s  %s", prefix, line)
            LOG.trace("%sHierarchy:", prefix)
            for line in node.pformat().splitlines():
                LOG.trace("%s  %s", prefix, line)

    def _pre_compile(self):
        """Called before the compilation of the root starts."""
        self._history.clear()
        self._level = 0

    def _post_compile(self, graph, node):
        """Called after the compilation of the root finishes successfully."""
        self._history.clear()
        self._level = 0

    @fasteners.locked
    def compile(self):
        """Compiles the contained item into a compiled equivalent."""
        if self._compilation is None:
            self._pre_compile()
            try:
                graph, node = self._compile(self._root, parent=None)
            except Exception:
                with excutils.save_and_reraise_exception():
                    # Always clear the history, to avoid retaining junk
                    # in memory that isn't needed to be in memory if
                    # compilation fails...
                    self._history.clear()
            else:
                self._post_compile(graph, node)
                if self._freeze:
                    graph.freeze()
                    node.freeze()
                self._compilation = Compilation(graph, node)
        return self._compilation



#    Copyright (C) 2014 Yahoo! Inc. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections
import contextlib
import copy
import logging

from oslo_utils import reflection

LOG = logging.getLogger(__name__)


class Listener:
    """Immutable helper that represents a notification listener/target."""

    def __init__(self, callback, args=None, kwargs=None, details_filter=None):
        """Initialize members

        :param callback: callback function
        :param details_filter: a callback that will be called before the
                               actual callback that can be used to discard
                               the event (thus avoiding the invocation of
                               the actual callback)
        :param args: non-keyworded arguments
        :type args: list/iterable/tuple
        :param kwargs: key-value pair arguments
        :type kwargs: dictionary
        """
        self._callback = callback
        self._details_filter = details_filter
        if not args:
            self._args = ()
        else:
            if not isinstance(args, tuple):
                self._args = tuple(args)
            else:
                self._args = args
        if not kwargs:
            self._kwargs = {}
        else:
            self._kwargs = kwargs.copy()

    @property
    def callback(self):
        """Callback (can not be none) to call with event + details."""
        return self._callback

    @property
    def details_filter(self):
        """Callback (may be none) to call to discard events + details."""
        return self._details_filter

    @property
    def kwargs(self):
        """Dictionary of keyword arguments to use in future calls."""
        return self._kwargs.copy()

    @property
    def args(self):
        """Tuple of positional arguments to use in future calls."""
        return self._args

    def __call__(self, event_type, details):
        """Activate the target callback with the given event + details.

        NOTE(harlowja): if a details filter callback exists and it returns
        a falsey value when called with the provided ``details``, then the
        target callback will **not** be called.
        """
        if self._details_filter is not None:
            if not self._details_filter(details):
                return
        kwargs = self._kwargs.copy()
        kwargs['details'] = details
        self._callback(event_type, *self._args, **kwargs)

    def __repr__(self):
        repr_msg = "{} object at 0x{:x} calling into '{!r}'".format(
            reflection.get_class_name(self, fully_qualified=False),
            id(self), self._callback)
        if self._details_filter is not None:
            repr_msg += " using details filter '%r'" % self._details_filter
        return "<%s>" % repr_msg

    def is_equivalent(self, callback, details_filter=None):
        """Check if the callback is same

        :param callback: callback used for comparison
        :param details_filter: callback used for comparison
        :returns: false if not the same callback, otherwise true
        :rtype: boolean
        """
        if not reflection.is_same_callback(self._callback, callback):
            return False
        if details_filter is not None:
            if self._details_filter is None:
                return False
            else:
                return reflection.is_same_callback(self._details_filter,
                                                   details_filter)
        else:
            return self._details_filter is None

    def __eq__(self, other):
        if isinstance(other, Listener):
            return self.is_equivalent(other._callback,
                                      details_filter=other._details_filter)
        else:
            return NotImplemented

    def __ne__(self, other):
        return not self.__eq__(other)


class Notifier:
    """A notification (`pub/sub`_ *like*) helper class.

    It is intended to be used to subscribe to notifications of events
    occurring as well as allow a entity to post said notifications to any
    associated subscribers without having either entity care about how this
    notification occurs.

    **Not** thread-safe when a single notifier is mutated at the same
    time by multiple threads. For example having multiple threads call
    into :py:meth:`.register` or :py:meth:`.reset` at the same time could
    potentially end badly. It is thread-safe when
    only :py:meth:`.notify` calls or other read-only actions (like calling
    into :py:meth:`.is_registered`) are occurring at the same time.

    .. _pub/sub: http://en.wikipedia.org/wiki/Publish%E2%80%93subscribe_pattern
    """

    #: Keys that can *not* be used in callbacks arguments
    RESERVED_KEYS = ('details',)

    #: Kleene star constant that is used to receive all notifications
    ANY = '*'

    #: Events which can *not* be used to trigger notifications
    _DISALLOWED_NOTIFICATION_EVENTS = {ANY}

    def __init__(self):
        self._topics = collections.defaultdict(list)

    def __len__(self):
        """Returns how many callbacks are registered.

        :returns: count of how many callbacks are registered
        :rtype: number
        """
        count = 0
        for (_event_type, listeners) in self._topics.items():
            count += len(listeners)
        return count

    def is_registered(self, event_type, callback, details_filter=None):
        """Check if a callback is registered.

        :returns: checks if the callback is registered
        :rtype: boolean
        """
        for listener in self._topics.get(event_type, []):
            if listener.is_equivalent(callback, details_filter=details_filter):
                return True
        return False

    def reset(self):
        """Forget all previously registered callbacks."""
        self._topics.clear()

    def notify(self, event_type, details):
        """Notify about event occurrence.

        All callbacks registered to receive notifications about given
        event type will be called. If the provided event type can not be
        used to emit notifications (this is checked via
        the :meth:`.can_be_registered` method) then it will silently be
        dropped (notification failures are not allowed to cause or
        raise exceptions).

        :param event_type: event type that occurred
        :param details: additional event details *dictionary* passed to
                        callback keyword argument with the same name
        :type details: dictionary
        """
        if not self.can_trigger_notification(event_type):
            LOG.debug("Event type '%s' is not allowed to trigger"
                      " notifications", event_type)
            return
        listeners = list(self._topics.get(self.ANY, []))
        listeners.extend(self._topics.get(event_type, []))
        if not listeners:
            return
        if not details:
            details = {}
        for listener in listeners:
            try:
                listener(event_type, details.copy())
            except Exception:
                LOG.warning("Failure calling listener %s to notify about event"
                            " %s, details: %s", listener, event_type,
                            details, exc_info=True)

    def register(self, event_type, callback,
                 args=None, kwargs=None, details_filter=None):
        """Register a callback to be called when event of a given type occurs.

        Callback will be called with provided ``args`` and ``kwargs`` and
        when event type occurs (or on any event if ``event_type`` equals to
        :attr:`.ANY`). It will also get additional keyword argument,
        ``details``, that will hold event details provided to the
        :meth:`.notify` method (if a details filter callback is provided then
        the target callback will *only* be triggered if the details filter
        callback returns a truthy value).

        :param event_type: event type input
        :param callback: function callback to be registered.
        :param args: non-keyworded arguments
        :type args: list
        :param kwargs: key-value pair arguments
        :type kwargs: dictionary
        """
        if not callable(callback):
            raise ValueError("Event callback must be callable")
        if details_filter is not None:
            if not callable(details_filter):
                raise ValueError("Details filter must be callable")
        if not self.can_be_registered(event_type):
            raise ValueError("Disallowed event type '%s' can not have a"
                             " callback registered" % event_type)
        if self.is_registered(event_type, callback,
                              details_filter=details_filter):
            raise ValueError("Event callback already registered with"
                             " equivalent details filter")
        if kwargs:
            for k in self.RESERVED_KEYS:
                if k in kwargs:
                    raise KeyError("Reserved key '%s' not allowed in "
                                   "kwargs" % k)
        self._topics[event_type].append(
            Listener(callback,
                     args=args, kwargs=kwargs,
                     details_filter=details_filter))

    def deregister(self, event_type, callback, details_filter=None):
        """Remove a single listener bound to event ``event_type``.

        :param event_type: deregister listener bound to event_type
        """
        if event_type not in self._topics:
            return False
        for i, listener in enumerate(self._topics.get(event_type, [])):
            if listener.is_equivalent(callback, details_filter=details_filter):
                self._topics[event_type].pop(i)
                return True
        return False

    def deregister_event(self, event_type):
        """Remove a group of listeners bound to event ``event_type``.

        :param event_type: deregister listeners bound to event_type
        """
        return len(self._topics.pop(event_type, []))

    def copy(self):
        c = copy.copy(self)
        c._topics = collections.defaultdict(list)
        for (event_type, listeners) in self._topics.items():
            c._topics[event_type] = listeners[:]
        return c

    def listeners_iter(self):
        """Return an iterator over the mapping of event => listeners bound.

        NOTE(harlowja): Each listener in the yielded (event, listeners)
        tuple is an instance of the :py:class:`~.Listener`  type, which
        itself wraps a provided callback (and its details filter
        callback, if any).
        """
        for event_type, listeners in self._topics.items():
            if listeners:
                yield (event_type, listeners)

    def can_be_registered(self, event_type):
        """Checks if the event can be registered/subscribed to."""
        return True

    def can_trigger_notification(self, event_type):
        """Checks if the event can trigger a notification.

        :param event_type: event that needs to be verified
        :returns: whether the event can trigger a notification
        :rtype: boolean
        """
        if event_type in self._DISALLOWED_NOTIFICATION_EVENTS:
            return False
        else:
            return True


class RestrictedNotifier(Notifier):
    """A notification class that restricts events registered/triggered.

    NOTE(harlowja): This class unlike :class:`.Notifier` restricts and
    disallows registering callbacks for event types that are not declared
    when constructing the notifier.
    """

    def __init__(self, watchable_events, allow_any=True):
        super().__init__()
        self._watchable_events = frozenset(watchable_events)
        self._allow_any = allow_any

    def events_iter(self):
        """Returns iterator of events that can be registered/subscribed to.

        NOTE(harlowja): does not include back the ``ANY`` event type as that
        meta-type is not a specific event but is a capture-all that does not
        imply the same meaning as specific event types.
        """
        yield from self._watchable_events

    def can_be_registered(self, event_type):
        """Checks if the event can be registered/subscribed to.

        :param event_type: event that needs to be verified
        :returns: whether the event can be registered/subscribed to
        :rtype: boolean
        """
        return (event_type in self._watchable_events or
                (event_type == self.ANY and self._allow_any))


@contextlib.contextmanager
def register_deregister(notifier, event_type, callback=None,
                        args=None, kwargs=None, details_filter=None):
    """Context manager that registers a callback, then deregisters on exit.

    NOTE(harlowja): if the callback is none, then this registers nothing, which
                    is different from the behavior of the ``register`` method
                    which will *not* accept none as it is not callable...
    """
    if callback is None:
        yield
    else:
        notifier.register(event_type, callback,
                          args=args, kwargs=kwargs,
                          details_filter=details_filter)
        try:
            yield
        finally:
            notifier.deregister(event_type, callback,
                                details_filter=details_filter)



#    Copyright (C) 2014 Yahoo! Inc. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_utils import uuidutils

from taskflow.engines.action_engine import executor
from taskflow.engines.worker_based import protocol as pr
from taskflow import exceptions as excp
from taskflow import test
from taskflow.test import mock
from taskflow.tests import utils
from taskflow.types import failure


class Unserializable:
    pass


class TestProtocolValidation(test.TestCase):
    def test_send_notify(self):
        msg = pr.Notify()
        pr.Notify.validate(msg.to_dict(), False)

    def test_send_notify_invalid(self):
        msg = {
            'all your base': 'are belong to us',
        }
        self.assertRaises(excp.InvalidFormat,
                          pr.Notify.validate, msg, False)

    def test_reply_notify(self):
        msg = pr.Notify(topic="bob", tasks=['a', 'b', 'c'])
        pr.Notify.validate(msg.to_dict(), True)

    def test_reply_notify_invalid(self):
        msg = {
            'topic': {},
            'tasks': 'not yours',
        }
        self.assertRaises(excp.InvalidFormat,
                          pr.Notify.validate, msg, True)

    def test_request(self):
        request = pr.Request(utils.DummyTask("hi"),
                             uuidutils.generate_uuid(),
                             pr.EXECUTE, {}, 1.0)
        pr.Request.validate(request.to_dict())

    def test_request_invalid(self):
        msg = {
            'task_name': 1,
            'task_cls': False,
            'arguments': [],
        }
        self.assertRaises(excp.InvalidFormat, pr.Request.validate, msg)

    def test_request_invalid_action(self):
        request = pr.Request(utils.DummyTask("hi"),
                             uuidutils.generate_uuid(),
                             pr.EXECUTE, {}, 1.0)
        request = request.to_dict()
        request['action'] = 'NOTHING'
        self.assertRaises(excp.InvalidFormat, pr.Request.validate, request)

    def test_response_progress(self):
        msg = pr.Response(pr.EVENT, details={'progress': 0.5},
                          event_type='blah')
        pr.Response.validate(msg.to_dict())

    def test_response_completion(self):
        msg = pr.Response(pr.SUCCESS, result=1)
        pr.Response.validate(msg.to_dict())

    def test_response_mixed_invalid(self):
        msg = pr.Response(pr.EVENT,
                          details={'progress': 0.5},
                          event_type='blah', result=1)
        self.assertRaises(excp.InvalidFormat, pr.Response.validate, msg)

    def test_response_bad_state(self):
        msg = pr.Response('STUFF')
        self.assertRaises(excp.InvalidFormat, pr.Response.validate, msg)


class TestProtocol(test.TestCase):

    def setUp(self):
        super().setUp()
        self.task = utils.DummyTask()
        self.task_uuid = 'task-uuid'
        self.task_action = 'execute'
        self.task_args = {'a': 'a'}
        self.timeout = 60

    def request(self, **kwargs):
        request_kwargs = dict(task=self.task,
                              uuid=self.task_uuid,
                              action=self.task_action,
                              arguments=self.task_args,
                              timeout=self.timeout)
        request_kwargs.update(kwargs)
        return pr.Request(**request_kwargs)

    def request_to_dict(self, **kwargs):
        to_dict = dict(task_cls=self.task.name,
                       task_name=self.task.name,
                       task_version=self.task.version,
                       action=self.task_action,
                       arguments=self.task_args)
        to_dict.update(kwargs)
        return to_dict

    def test_request_transitions(self):
        request = self.request()
        self.assertEqual(pr.WAITING, request.current_state)
        self.assertIn(request.current_state, pr.WAITING_STATES)
        self.assertRaises(excp.InvalidState, request.transition, pr.SUCCESS)
        self.assertFalse(request.transition(pr.WAITING))
        self.assertTrue(request.transition(pr.PENDING))
        self.assertTrue(request.transition(pr.RUNNING))
        self.assertTrue(request.transition(pr.SUCCESS))
        for s in (pr.PENDING, pr.WAITING):
            self.assertRaises(excp.InvalidState, request.transition, s)

    def test_creation(self):
        request = self.request()
        self.assertEqual(self.task_uuid, request.uuid)
        self.assertEqual(self.task, request.task)
        self.assertFalse(request.future.done())

    def test_to_dict_default(self):
        request = self.request()
        self.assertEqual(self.request_to_dict(), request.to_dict())

    def test_to_dict_with_result(self):
        request = self.request(result=333)
        self.assertEqual(self.request_to_dict(result=('success', 333)),
                         request.to_dict())

    def test_to_dict_with_result_none(self):
        request = self.request(result=None)
        self.assertEqual(self.request_to_dict(result=('success', None)),
                         request.to_dict())

    def test_to_dict_with_result_failure(self):
        a_failure = failure.Failure.from_exception(RuntimeError('Woot!'))
        expected = self.request_to_dict(result=('failure',
                                                a_failure.to_dict()))
        request = self.request(result=a_failure)
        self.assertEqual(expected, request.to_dict())

    def test_to_dict_with_failures(self):
        a_failure = failure.Failure.from_exception(RuntimeError('Woot!'))
        request = self.request(failures={self.task.name: a_failure})
        expected = self.request_to_dict(
            failures={self.task.name: a_failure.to_dict()})
        self.assertEqual(expected, request.to_dict())

    def test_to_dict_with_invalid_json_failures(self):
        exc = RuntimeError(Unserializable())
        a_failure = failure.Failure.from_exception(exc)
        request = self.request(failures={self.task.name: a_failure})
        expected = self.request_to_dict(
            failures={self.task.name: a_failure.to_dict(include_args=False)})
        self.assertEqual(expected, request.to_dict())

    @mock.patch('oslo_utils.timeutils.now')
    def test_pending_not_expired(self, now):
        now.return_value = 0
        request = self.request()
        now.return_value = self.timeout - 1
        self.assertFalse(request.expired)

    @mock.patch('oslo_utils.timeutils.now')
    def test_pending_expired(self, now):
        now.return_value = 0
        request = self.request()
        now.return_value = self.timeout + 1
        self.assertTrue(request.expired)

    @mock.patch('oslo_utils.timeutils.now')
    def test_running_not_expired(self, now):
        now.return_value = 0
        request = self.request()
        request.transition(pr.PENDING)
        request.transition(pr.RUNNING)
        now.return_value = self.timeout + 1
        self.assertFalse(request.expired)

    def test_set_result(self):
        request = self.request()
        request.set_result(111)
        result = request.future.result()
        self.assertEqual((executor.EXECUTED, 111), result)



#    Copyright (C) 2014 Yahoo! Inc. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import threading

from oslo_utils import uuidutils

from taskflow.engines.worker_based import dispatcher
from taskflow.engines.worker_based import protocol as pr
from taskflow.engines.worker_based import proxy
from taskflow import test
from taskflow.test import mock
from taskflow.tests import utils as test_utils
from taskflow.types import latch
from taskflow.utils import threading_utils

TEST_EXCHANGE, TEST_TOPIC = ('test-exchange', 'test-topic')
POLLING_INTERVAL = 0.01


class TestMessagePump(test.TestCase):
    def test_notify(self):
        barrier = threading.Event()

        on_notify = mock.MagicMock()
        on_notify.side_effect = lambda *args, **kwargs: barrier.set()

        handlers = {pr.NOTIFY: dispatcher.Handler(on_notify)}
        p = proxy.Proxy(TEST_TOPIC, TEST_EXCHANGE, handlers,
                        transport='memory',
                        transport_options={
                            'polling_interval': POLLING_INTERVAL,
                        })

        t = threading_utils.daemon_thread(p.start)
        t.start()
        p.wait()
        p.publish(pr.Notify(), TEST_TOPIC)

        self.assertTrue(barrier.wait(test_utils.WAIT_TIMEOUT))
        p.stop()
        t.join()

        self.assertTrue(on_notify.called)
        on_notify.assert_called_with({}, mock.ANY)

    def test_response(self):
        barrier = threading.Event()

        on_response = mock.MagicMock()
        on_response.side_effect = lambda *args, **kwargs: barrier.set()

        handlers = {pr.RESPONSE: dispatcher.Handler(on_response)}
        p = proxy.Proxy(TEST_TOPIC, TEST_EXCHANGE, handlers,
                        transport='memory',
                        transport_options={
                            'polling_interval': POLLING_INTERVAL,
                        })

        t = threading_utils.daemon_thread(p.start)
        t.start()
        p.wait()
        resp = pr.Response(pr.RUNNING)
        p.publish(resp, TEST_TOPIC)

        self.assertTrue(barrier.wait(test_utils.WAIT_TIMEOUT))
        self.assertTrue(barrier.is_set())
        p.stop()
        t.join()

        self.assertTrue(on_response.called)
        on_response.assert_called_with(resp.to_dict(), mock.ANY)

    def test_multi_message(self):
        message_count = 30
        barrier = latch.Latch(message_count)
        countdown = lambda data, message: barrier.countdown()

        on_notify = mock.MagicMock()
        on_notify.side_effect = countdown

        on_response = mock.MagicMock()
        on_response.side_effect = countdown

        on_request = mock.MagicMock()
        on_request.side_effect = countdown

        handlers = {
            pr.NOTIFY: dispatcher.Handler(on_notify),
            pr.RESPONSE: dispatcher.Handler(on_response),
            pr.REQUEST: dispatcher.Handler(on_request),
        }
        p = proxy.Proxy(TEST_TOPIC, TEST_EXCHANGE, handlers,
                        transport='memory',
                        transport_options={
                            'polling_interval': POLLING_INTERVAL,
                        })

        t = threading_utils.daemon_thread(p.start)
        t.start()
        p.wait()

        for i in range(0, message_count):
            j = i % 3
            if j == 0:
                p.publish(pr.Notify(), TEST_TOPIC)
            elif j == 1:
                p.publish(pr.Response(pr.RUNNING), TEST_TOPIC)
            else:
                p.publish(pr.Request(test_utils.DummyTask("dummy_%s" % i),
                                     uuidutils.generate_uuid(),
                                     pr.EXECUTE, [], None), TEST_TOPIC)

        self.assertTrue(barrier.wait(test_utils.WAIT_TIMEOUT))
        self.assertEqual(0, barrier.needed)
        p.stop()
        t.join()

        self.assertTrue(on_notify.called)
        self.assertTrue(on_response.called)
        self.assertTrue(on_request.called)

        self.assertEqual(10, on_notify.call_count)
        self.assertEqual(10, on_response.call_count)
        self.assertEqual(10, on_request.call_count)

        call_count = sum([
            on_notify.call_count,
            on_response.call_count,
            on_request.call_count,
        ])
        self.assertEqual(message_count, call_count)
